//-----------------------------------------------------------------------------
// Copyright Notice
//
//   Copyright 2002 Sandia Corporation. Under the terms
//   of Contract DE-AC04-94AL85000 with Sandia Corporation, the U.S.
//   Government retains certain rights in this software.
//
//    Xyce(TM) Parallel Electrical Simulator
//    Copyright (C) 2002-2016 Sandia Corporation
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//-----------------------------------------------------------------------------

//-------------------------------------------------------------------------
//
// Purpose        : Output Manager
//
// Special Notes  :
//
// Creator        : Robert J. Hoekstra, SNL, Parallel Computational Sciences
//
// Creation Date  : 10/10/00
//
//
//
//
//-------------------------------------------------------------------------

#include <Xyce_config.h>

#include <iostream>
#include <fstream>
#include <sstream>

#include <N_DEV_DeviceMgr.h>
#include <N_DEV_DeviceBlock.h>
#include <N_DEV_DeviceSensitivities.h>
#include <N_ERH_ErrorMgr.h>
#include <N_IO_CircuitBlock.h>
#include <N_IO_CmdParse.h>
#include <N_IO_InitialConditions.h>
#include <N_IO_Op.h>
#include <N_IO_OpBuilders.h>
#include <N_IO_OptionBlock.h>
#include <N_IO_OutputFileBase.h>
#include <N_IO_OutputMgr.h>
#include <N_IO_OutputterTimePrn.h>
#include <N_IO_OutputterHomotopy.h>
#include <N_IO_OutputterDC.h>
#include <N_IO_OutputterAC.h>
#include <N_IO_OutputterNoise.h>
#include <N_IO_OutputterHB.h>
#include <N_IO_OutputterMPDE.h>
#include <N_IO_OutputterTransient.h>
#include <N_IO_OutputterSensitivity.h>
#include <N_IO_OutputterRawOverride.h>
#include <N_IO_PkgOptionsMgr.h>
#include <N_IO_SpiceSeparatedFieldTool.h>
#include <N_IO_mmio.h>
#include <N_LAS_BlockVector.h>
#include <N_LAS_Matrix.h>
#include <N_LAS_System.h>
#include <N_LAS_Vector.h>
#include <N_PDS_Comm.h>
#include <N_PDS_Serial.h>
#include <N_PDS_MPI.h>
#include <N_TOP_Topology.h>
#include <N_UTL_Algorithm.h>
#include <N_UTL_FeatureTest.h>
#include <N_UTL_LogStream.h>
#include <N_UTL_Marshal.h>
#include <N_UTL_Math.h>
#include <N_UTL_ExtendedString.h>
#include <N_UTL_Version.h>

namespace Xyce {
namespace IO {

namespace SensitivityOptions {

enum {
  DIRECT   = 0x01,
  ADJOINT  = 0x02,
  SCALED   = 0x04,
  UNSCALED = 0x08
};

}

//-----------------------------------------------------------------------------
// Function      : OutputMgr::OutputMgr
// Purpose       : constructor
// Special Notes :
// Scope         : public
// Creator       : Robert Hoekstra, SNL, Parallel Computational Sciences
// Creation Date : 10/10/00
//-----------------------------------------------------------------------------
OutputMgr::OutputMgr(
  const CmdParse &              command_line,
  Util::Op::BuilderManager &    op_builder_manager,
  const Topo::Topology &        topology)
  : title_(command_line.getArgumentValue("netlist")),
    netlistFilename_(command_line.getArgumentValue("netlist")),
    filenameSuffix_(),
    opBuilderManager_(op_builder_manager),
    topology_(topology),
    enableHomotopyFlag_(false),
    enableSensitivityFlag_(false),
    adjointSensitivityFlag_(false),
    sensitivityOptions_(0),
    pts_per_summary_(-1),
    pts_per_summary_Given(false),
    PRINTdcstart_(0.0),
    PRINTdcstop_(0.0),
    PRINTdcvalue_(0.0),
    PRINTdcname_(""),
    initialOutputInterval_(0.0),
    printHeader_(true),
    printFooter_(true),
    outputVersionInRawFile_(false),
    outputCalledBefore_(false),
    dcLoopNumber_(0),
    maxDCSteps_(0),
    hdf5FileNameGiven_(false),
    hdf5HeaderWritten_(false),
    hdf5IndexValue_(0)
{
  if (command_line.getArgumentValue("-delim") == "TAB")
  {
    defaultPrintParameters_.delimiter_ = "\t";
  }
  else if (command_line.getArgumentValue("-delim") == "COMMA")
  {
    defaultPrintParameters_.delimiter_ = ",";
  }
  else
  {
    defaultPrintParameters_.delimiter_ = command_line.getArgumentValue("-delim");
  }

  if (command_line.argExists("-a"))
  {
    defaultPrintParameters_.asciiRaw_ = true;
  }

  if (command_line.argExists("-r")) 
  {
    defaultPrintParameters_.overrideRaw_ = true;
    defaultPrintParameters_.filename_ = command_line.getArgumentValue("-r");
    defaultPrintParameters_.format_ = 
      defaultPrintParameters_.asciiRaw_ ? Format::RAW_ASCII : Format::RAW;
  }

  // If the output file is specified on the command line it takes precedence
  else if (command_line.argExists("-o"))
  {
    defaultPrintParameters_.filename_ = command_line.getArgumentValue("-o");
  }

  sensitivityOptions_ |= SensitivityOptions::ADJOINT;
  sensitivityOptions_ |= SensitivityOptions::UNSCALED;
}

//-----------------------------------------------------------------------------
// Function      : OutputMgr::~OutputMgr
// Purpose       : destructor
// Special Notes :
// Scope         : public
// Creator       : Robert Hoekstra, SNL, Parallel Computational Sciences
// Creation Date : 10/10/00
//-----------------------------------------------------------------------------
OutputMgr::~OutputMgr()
{
  OutputterMap::iterator it = outputterMap_.begin();
  for ( ; it != outputterMap_.end(); ++it)
  {
    OutputterMap::mapped_type::iterator it2 = (*it).second.begin();
    for ( ; it2 != (*it).second.end(); ++it2)
    {
      delete (*it2);
    }
  }

  OpenPathStreamMap::iterator it1 = openPathStreamMap_.begin();
  for ( ; it1 != openPathStreamMap_.end(); ++it1)
  {
    delete (*it1).second.second;
  }
}

//-----------------------------------------------------------------------------
// Function      : OutputMgr::getVarTypes
// Purpose       :
// Special Notes :
// Scope         :
// Creator       :
// Creation Date :
//-----------------------------------------------------------------------------
const std::vector<char> & OutputMgr::getVarTypes() const
{
  return topology_.getVarTypes();
}

//-----------------------------------------------------------------------------
// Function      : OutputMgr::getSolutionNodeMap
// Purpose       :
// Special Notes :
// Scope         :
// Creator       :
// Creation Date :
//-----------------------------------------------------------------------------
const NodeNameMap & OutputMgr::getSolutionNodeMap() const
{
  return topology_.getSolutionNodeNameMap();
}

//-----------------------------------------------------------------------------
// Function      : OutputMgr::getStateNodeMap
// Purpose       :
// Special Notes :
// Scope         :
// Creator       :
// Creation Date :
//-----------------------------------------------------------------------------
const NodeNameMap & OutputMgr::getStateNodeMap() const
{
  return topology_.getStateNodeNameMap();
}

//-----------------------------------------------------------------------------
// Function      : OutputMgr::getStoreNodeMap
// Purpose       :
// Special Notes :
// Scope         :
// Creator       :
// Creation Date :
//-----------------------------------------------------------------------------
const NodeNameMap & OutputMgr::getStoreNodeMap() const
{
  return topology_.getStoreNodeNameMap();
}

//-----------------------------------------------------------------------------
// Function      : OutputMgr::getExternalNodeMap
// Purpose       :
// Special Notes :
// Scope         :
// Creator       :
// Creation Date :
//-----------------------------------------------------------------------------
const NodeNameMap & OutputMgr::getExternalNodeMap() const
{
  return topology_.getExternalNodeNameMap();
}

//-----------------------------------------------------------------------------
// Function      : OutputMgr::getBranchVarsNodeMap
// Purpose       :
// Special Notes :
// Scope         :
// Creator       :
// Creation Date :
//-----------------------------------------------------------------------------
const NodeNameMap & OutputMgr::getBranchVarsNodeMap() const
{
  return topology_.getBranchVarsNodeNameMap();
}

//-----------------------------------------------------------------------------
// Function      : OutputMgr::notify
// Purpose       :
// Special Notes :
// Scope         :
// Creator       :
// Creation Date :
//-----------------------------------------------------------------------------
void OutputMgr::notify(
  const Analysis::StepEvent &   step_event)
{
  switch (step_event.state_) 
  {
    case Analysis::StepEvent::INITIALIZE:
      outputState_.stepMaxCount_ = step_event.count_;
      break;

    case Analysis::StepEvent::STEP_STARTED:
      outputState_.stepLoopNumber_ = step_event.count_;
      startStep(outputState_.stepLoopNumber_, outputState_.stepMaxCount_);
      break;

    case Analysis::StepEvent::STEP_COMPLETED:
      break;

    case Analysis::StepEvent::FINISH:
      steppingComplete();
      break;
  }
}

//-----------------------------------------------------------------------------
// Function      : OutputMgr::openFile
// Purpose       : open named file in given mode, create stream
// Special Notes :
// Scope         : public
// Creator       : David Baur, Raytheon
// Creation Date : 07/08/2013
//-----------------------------------------------------------------------------
std::ostream * OutputMgr::openFile(
  const std::string &           path,
  std::ios_base::openmode       mode)
{
  OpenPathStreamMap::iterator it= openPathStreamMap_.find(path);

  if (path == "CONSOLE")
  {
    return &Xyce::dout();
  }
  else if (it != openPathStreamMap_.end()) 
  {
    ++(*it).second.first;
    return (*it).second.second;
  }
  else 
  {
    std::ostream *os = new std::ofstream(path.c_str(), mode);
    openPathStreamMap_[path] = std::pair<int, std::ostream *>(1, os);

    if (!os->good())
    {
      Report::UserFatal() << "Failure opening " << path;
    }

    return os;
  }
}

//-----------------------------------------------------------------------------
// Function      : OutputMgr::openFile
// Purpose       : open named file for output only, create stream
// Special Notes :
// Scope         : public
// Creator       : David Baur, Raytheon
// Creation Date : 07/08/2013
//-----------------------------------------------------------------------------
std::ostream * OutputMgr::openFile(const std::string & path)
{
  return openFile(path, std::ios_base::out);
}

//-----------------------------------------------------------------------------
// Function      : OutputMgr::openBinaryFile
// Purpose       : open named file in binary mode for output only, create
//                 stream
// Special Notes :
// Scope         : public
// Creator       : David Baur, Raytheon
// Creation Date : 07/08/2013
//-----------------------------------------------------------------------------
std::ostream * OutputMgr::openBinaryFile(const std::string & path)
{
  return openFile(path, std::ios_base::out | std::ios_base::binary);
}

//-----------------------------------------------------------------------------
// Function      : OutputMgr::closeFile
// Purpose       : Close given stream
// Special Notes :
// Scope         : public
// Creator       : David Baur, Raytheon
// Creation Date : 07/08/2013
//-----------------------------------------------------------------------------
int OutputMgr::closeFile(std::ostream * os)
{
  if (os == &Xyce::dout())
  {
    return 1;
  }

  int open_count= 0;

  for (OpenPathStreamMap::iterator it= openPathStreamMap_.begin(); it != openPathStreamMap_.end(); ++it) 
  {
    if ((*it).second.second == os) 
    {
      open_count= --(*it).second.first;
      if (open_count == 0) 
      {
        delete os;
        openPathStreamMap_.erase(it);
        break;
      }
    }
  }

  return open_count;
}


namespace {

//-----------------------------------------------------------------------------
// Function      : testAndSet
// Purpose       : test if set contains element and then add the element
// Special Notes :
// Scope         : public
// Creator       : David G. Baur  Raytheon  Sandia National Laboratories 1355
// Creation Date : Tue Jun 24 09:38:58 2014
//-----------------------------------------------------------------------------
///
/// Returns true of set contained element on entry, adds element if it was not found
///
/// @invariant set contains new element on completion
///
/// @param s Set to find element
/// @param t Element to find
///
/// @return true if set contained element before insertion
///
template<class S, class T>
inline bool testAndSet(S &s, const T &t)
{
  bool found = s.find(t) != s.end();
  if (!found)
    s.insert(t);
  return found;
}

}

//-----------------------------------------------------------------------------
// Function      : OutputMgr::prepareOutput
// Purpose       : checks to make sure requested values can be output
// Special Notes : Primary purpose is for error checking BEFORE calculation.
//                 Also allows any name translation for output, if needed.
//
//                 This function also now serves as an initialization for
//                 allNodes_.  allNodes_ needs to be set up after topology
//                 is done setting up.
//
// erkeite  8/10/2007:
// Note that this is currently(august 2007) one of 3 different functions
// that check the .print line for errors.  That is a bit confusing,
// unfortunately.  For commentary and some explanation of this function,
// as well as the other two, see the comments for the delayedPrintLineDiagnostics
// function.  This whole functionality really should be refactored, or
// at least made more transparent to other developers.
//
// Function renamed from check_output to prepareOutput and heavily reorganized
// by David Baur on 6/28/2013.
//
// Scope         : public
// Creator       : Dave Shirley, PSSI
// Creation Date : 06/01/05
//-----------------------------------------------------------------------------
void OutputMgr::prepareOutput(
  Parallel::Machine             comm,
  Analysis::Mode                analysis_mode)
{
  // Setup rawfile if requested
  if (defaultPrintParameters_.overrideRaw_)
  {
    Outputter::enableRawOverrideOutput(comm, *this, analysis_mode);

    if (activeOutputterStack_.empty())
      activeOutputterStack_.push_back(std::vector<Outputter::Interface *>());

    addActiveOutputter(PrintType::RAW_OVERRIDE, analysis_mode);
  }
  else
  {
    switch (analysis_mode)
    {
      case Analysis::ANP_MODE_INVALID:
      case Analysis::ANP_MODE_DC_OP:
      case Analysis::ANP_MODE_MOR:
        break;

      case Analysis::ANP_MODE_DC_SWEEP:
        if (!testAndSet(enabledAnalysisSet_, Analysis::ANP_MODE_DC_SWEEP))
          Outputter::enableDCOutput(comm, *this, analysis_mode);
        addActiveOutputter(PrintType::TRAN, analysis_mode);
        break;

      case Analysis::ANP_MODE_TRANSIENT:
        if (!testAndSet(enabledAnalysisSet_, Analysis::ANP_MODE_TRANSIENT))
          Outputter::enableTransientOutput(comm, *this, analysis_mode);
        addActiveOutputter(PrintType::TRAN, analysis_mode);
        break;

      case Analysis::ANP_MODE_MPDE:
        if (!testAndSet(enabledAnalysisSet_, Analysis::ANP_MODE_MPDE))
          Outputter::enableMPDEOutput(comm, *this, analysis_mode);
        addActiveOutputter(PrintType::MPDE, analysis_mode);
        addActiveOutputter(PrintType::MPDE_IC, analysis_mode);
        addActiveOutputter(PrintType::MPDE_STARTUP, analysis_mode);
        break;

      case Analysis::ANP_MODE_HB:
        if (!testAndSet(enabledAnalysisSet_, Analysis::ANP_MODE_HB))
          Outputter::enableHBOutput(comm, *this, analysis_mode);
        addActiveOutputter(PrintType::HB, analysis_mode);
        addActiveOutputter(PrintType::HB_IC, analysis_mode);
        addActiveOutputter(PrintType::HB_STARTUP, analysis_mode);
        break;

      case Analysis::ANP_MODE_AC:
        if (!testAndSet(enabledAnalysisSet_, Analysis::ANP_MODE_AC))
          Outputter::enableACOutput(comm, *this, analysis_mode);
        addActiveOutputter(PrintType::AC, analysis_mode);
        addActiveOutputter(PrintType::AC_IC, analysis_mode);
        break;
      case Analysis::ANP_MODE_NOISE:
        if (!testAndSet(enabledAnalysisSet_, Analysis::ANP_MODE_NOISE))
          Outputter::enableNoiseOutput(comm, *this, analysis_mode);
        addActiveOutputter(PrintType::NOISE, analysis_mode);
        break;
    }

    if (enableHomotopyFlag_)
    {
      if (!testAndSet(enabledAnalysisSet_, (Analysis::Mode) (Analysis::ANP_MODE_INVALID + 100)))
        Outputter::enableHomotopyOutput(comm, *this, analysis_mode);
      addActiveOutputter(PrintType::HOMOTOPY, analysis_mode);
    }

    if (enableSensitivityFlag_)
    {
      if (!testAndSet(enabledAnalysisSet_, (Analysis::Mode) (Analysis::ANP_MODE_INVALID + 101)))
        Outputter::enableSensitivityOutput(comm, *this, analysis_mode);
      addActiveOutputter(PrintType::SENS, analysis_mode);

      // allocate an adjoint output object, but do not add to the active list.
      // Need to re-think the INVALID+102 thing.
      if ( adjointSensitivityFlag_ )
      {
        if (!testAndSet(enabledAnalysisSet_, (Analysis::Mode) (Analysis::ANP_MODE_INVALID + 102)))
          Outputter::enableAdjointSensitivityOutput(comm, *this, analysis_mode);

        // ERK.  NOT putting the transient adjoint outputters on the stack.  
        // Accessing them (for now) directly from the outputterMap.  I am punting 
        // on making the active outputter structure work for transient adjoints.
      }
    }
  }

  if (outputterMap_.empty())
  {
    Outputter::TimePrn *outputter = new Outputter::TimePrn(comm, *this, defaultPrintParameters_);
    outputterMap_[PrintType::TRAN].push_back(outputter);
  }
}

//-----------------------------------------------------------------------------
// Function      : OutputMgr::registerNonlinearOptions
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter, SNL
// Creation Date : 09/25/07
//-----------------------------------------------------------------------------
bool OutputMgr::registerNonlinearOptions(const Util::OptionBlock & OB)
{
  setEnableHomotopyFlag(true);

  return true;
}

//-----------------------------------------------------------------------------
// Function      : OutputMgr::registerOutputOptions
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Robert Hoekstra, SNL, Parallel Computational Sciences
// Creation Date : 8/30/01
//-----------------------------------------------------------------------------
bool OutputMgr::registerOutputOptions(const Util::OptionBlock & option_block)
{

  // Need to capture a variable length of output intervals.  The use case is:
  //
  // .OPTIONS OUTPUT INITIAL_INTERVAL= <interval> [<t0> <i0> [<t1> <i1>...]]
  //
  // additional options can appear before or after the INITIAL_INTERVAL as in
  //
  //.OPTIONS OUTPUT HDF5FILE= xxxx INITIAL_INTERVAL=<interval> [<t0> <i0> [<t1> <i1>...]]
  // or
  // .OPTIONS OUTPUT INITIAL_INTERVAL= <interval> [<t0> <i0> [<t1> <i1>...]] HDF5FILE=xxxx
  //

  for (Util::ParamList::const_iterator it = option_block.begin(), end = option_block.end(); it != end; )
  {
    if ((*it).tag() == "INITIAL_INTERVAL")
    {
      // start handling a list of intervals
      // set interval value
      initialOutputInterval_= (*it).getImmutableValue<double>();
      // look for optional time pairs.
      outputIntervalPairs_.clear();
      bool doneWithTimePairs= false;

      // need to point at next parameter
      ++it;

      while (it != end && !doneWithTimePairs)
      {
        if ((*it).tag() == "TIME")
        {
          double t = (*it).getImmutableValue<double>();
          ++it;

          double iv = 0.0;
          if (it != end) {
            iv = (*it).getImmutableValue<double>();
            ++it;
          }
          else
          {
            Report::UserError0() << "Must specific intervals in pairs";
          }
          
          outputIntervalPairs_.push_back(Interval(t, iv));
        }
        else
        {
          // didn't find a time pair so bail out of this loop.
          doneWithTimePairs = true;
        }
      }
    }

    else if ((*it).tag() == "HDF5FILENAME")
    {
    // look for other option tags
      hdf5FileNameGiven_= true;
      hdf5FileName_= (*it).stringValue();
      ++it;
    }
    else if ((*it).tag()=="PRINTHEADER")
    {
      // look for flag to turn off the header
      printHeader_ = (*it).getImmutableValue<bool>();
      ++it;
    }
    else if ((*it).tag()=="PRINTFOOTER")
    {
      // look for flag to turn off "End of Xyce(TM) Simulation" line
      printFooter_= (*it).getImmutableValue<bool>();
      ++it;
    }
    else if ((*it).tag()=="OUTPUTVERSIONINRAWFILE")
    {
      // look for flag to toggle output of version in header of RAW file
     outputVersionInRawFile_ = (*it).getImmutableValue<bool>();
     ++it;
    }
    else
    {
      // silently ignore?
      ++it;
    }
  }

  return true;
}

//-----------------------------------------------------------------------------
// Function      : OutputMgr::getVariableList
// Purpose       :
// Special Notes :
// Scope         : 
// Creator       : 
// Creation Date : 
//-----------------------------------------------------------------------------
Util::ParamList OutputMgr::getVariableList() const
{
  Util::ParamList parameter_list;

  for (OutputParameterMap::const_iterator it1 = outputParameterMap_.begin(), 
      end1 = outputParameterMap_.end(); it1 != end1; ++it1)
  {
    const OutputParameterMap::mapped_type &parameter_vector = (*it1).second;

    for (OutputParameterMap::mapped_type::const_iterator it2 = parameter_vector.begin(), 
        end2 = parameter_vector.end(); it2 != end2; ++it2)
    {
      const PrintParameters &print_parameters = (*it2);

      // Populate stringStat, nodeStat and instanceState with variable names from print line
      std::map<std::string, bool> stringStat;
      std::map<std::string, bool> nodeStat;
      std::map<std::string, bool> instanceStat;

      for (Util::ParamList::const_iterator it3 = print_parameters.variableList_.begin(), 
          end3 = print_parameters.variableList_.end(); it3 != end3; ++it3)
      {
        const Util::Param &parameter = (*it3);
        parameter_list.push_back(parameter);
      }
    }
  }

  return parameter_list;
}

//-----------------------------------------------------------------------------
// Function      : OutputMgr::getVariableNames
// Purpose       :
// Special Notes :
// Scope         : private
// Creator       : Todd Coffey
// Creation Date : 07/28/09
//-----------------------------------------------------------------------------
std::vector<std::string>
OutputMgr::getVariableNames()
{
  std::vector<std::string> name_list;

  for (OutputParameterMap::const_iterator it1 = outputParameterMap_.begin(), 
      end1 = outputParameterMap_.end(); it1 != end1; ++it1)
  {
    const OutputParameterMap::mapped_type &parameter_vector = (*it1).second;

    for (OutputParameterMap::mapped_type::const_iterator it2 = parameter_vector.begin(), 
        end2 = parameter_vector.end(); it2 != end2; ++it2)
    {
      const PrintParameters &print_parameters = (*it2);

      for (Util::ParamList::const_iterator it3 = print_parameters.variableList_.begin(), 
          end3 = print_parameters.variableList_.end(); it3 != end3; ++it3)
      {
        const Util::Param &parameter = (*it3);
        name_list.push_back(parameter.tag());
      }
    }
  }

  return name_list;
}

#if 0
// ERK.  These 2 structures are just used in the commented-out debug outputs
// below.  They will be removed soon.
const char * ModeChar[] = 
{
"ANP_MODE_INVALID",
"ANP_MODE_DC_OP",
"ANP_MODE_DC_SWEEP",
"ANP_MODE_DC_NLPOISSON",
"ANP_MODE_TRANSIENT",
"ANP_MODE_MPDE",
"ANP_MODE_HB",
"ANP_MODE_AC",
"ANP_MODE_NOISE",
"ANP_MODE_MOR"
};

const char * OutputTypeChar[] = {
 "DC",
 "TRAN",
 "AC",
 "AC_IC",
 "HB_FD",
 "HB_TD",
 "HB_IC",
 "HB_STARTUP",
 "DCOP",
 "HOMOTOPY",
 "MPDE",
 "MPDE_IC",
 "MPDE_STARTUP",
 "SENS",
 "TRANADJOINT",
 "NOISE"
};

#endif

//-----------------------------------------------------------------------------
// Function      : OutputMgr::checkPrintParameters
// Purpose       : 
// Special Notes : 
// Scope         : public
// Creator       : Dave Baur, Raytheon
// Creation Date : 6/23/2014
//-----------------------------------------------------------------------------
void OutputMgr::checkPrintParameters(
  Parallel::Machine                     comm,
  const Util::Op::BuilderManager &      op_builder_manager)
{
  Util::Op::OpList tempOpList;

  for (OutputParameterMap::const_iterator it = outputParameterMap_.begin(), 
      end = outputParameterMap_.end(); it != end; ++it)
  {
#if 0
    // ERK.  These commented-out debug outputs will be removed soon.
    const OutputType::OutputType outType = (*it).first;
    std::cout << "outputParameterMap key = " << OutputTypeChar[outType]  << std::endl;
#endif

    for (std::vector<PrintParameters>::const_iterator it2 = (*it).second.begin(), 
        end2 = (*it).second.end(); it2 != end2; ++it2)
    {
      PrintParameters print_parameters = (*it2);
      fixupPrintParameters(comm, print_parameters);

#if 0
      // ERK.  These commented-out debug outputs will be removed soon.
      std::cout << "print_parameters information:\t";
      std::cout << ModeChar[print_parameters.analysisMode_];
      if (print_parameters.fallback_)
      {
        std::cout << "\tfallback=true " << std::endl;
      }
      else
      {
        std::cout << "\tfallback=false" << std::endl;
      }

      Util::ParamList::const_iterator ii1 = print_parameters.variableList_.begin();
      Util::ParamList::const_iterator endii1 = print_parameters.variableList_.end();
      for ( ; ii1 != endii1; ++ii1)
      {
        const Util::Param & par = *ii1;
        std::cout << par.tag() << "\t" << par.stringValue() <<std::endl;
      }
#endif

      makeOps(comm, op_builder_manager, print_parameters.netlistLocation_,
              print_parameters.variableList_.begin(), 
              print_parameters.variableList_.end(), 
              std::back_inserter<Util::Op::OpList>(tempOpList));

    }
  }

  for (Util::Op::OpList::iterator it = tempOpList.begin(), end = tempOpList.end(); it != end; ++it)
  {
    delete *it;
  }

  if (hdf5FileNameGiven_)
  {
    prepareHDF5Output(comm);
  }
}

namespace {

// functor class for use with erase/remove_if idiom
struct IsFallback
{
  bool operator()(const PrintParameters &print_parameters)
  {
    return print_parameters.fallback_;
  }
};

// Functor class for use with find_if
struct PrintNamesMatch
{
  PrintNamesMatch(const std::string & name)
    :Name(name)
  {}

  bool operator()(const PrintParameters &print_parameters)
  {
    return print_parameters.filename_ == Name;
  }
  
  const std::string &Name;
};
}

//-----------------------------------------------------------------------------
// Function      : OutputMgr::addOutputParameters
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : David G. Baur  Raytheon  Sandia National Laboratories 1355
// Creation Date : Wed Jul 23 07:21:53 2014
//-----------------------------------------------------------------------------
///
/// Adds the print parameters to the output type as follows:
///
/// If the new print parameters is a fallback and no print parameters
/// have been provided then the print parameters are added.
///
/// If the new print is not a fallback (is specific), then the fallback
/// print parameters are removed and this one is added to the list.
///
/// If a prior print parameter set is found in the vector that matches
/// the file name and file format of the given print parameter object,
/// the variable list in the given object is appended to the existing
/// object rather than adding a new element to the vector.  In this way,
/// multiple print lines to the same file are the same as having one
/// large print line with an augmented variable list. 
///
/// @invariant
///
/// @param output_type          output type to add the print parameters too
/// @param print_parameters     print parameters to add
///
//-----------------------------------------------------------------------------
void OutputMgr::addOutputPrintParameters(
  OutputType::OutputType        output_type,
  const PrintParameters &       print_parameters)
{
  OutputParameterMap::mapped_type &print_parameters_vector = outputParameterMap_[output_type];

  // only add a fallback if there is no prior fallback 
  if (print_parameters.fallback_ && outputParameterMap_[output_type].empty())
  {
    print_parameters_vector.push_back(print_parameters);
  }
  else  // there is either already a fallback, or we are not a fallback
  {
    // if the new parameters are not a fallback, delete any fallbacks
    if (!print_parameters.fallback_)
      print_parameters_vector.erase(
         std::remove_if(print_parameters_vector.begin(),
                        print_parameters_vector.end(),
                        IsFallback()),
         print_parameters_vector.end());

    // Find if there's already a parameter list in the vector with the same
    // file name:
    OutputParameterMap::mapped_type::iterator it=std::find_if(
       print_parameters_vector.begin(),
       print_parameters_vector.end(),
       PrintNamesMatch(print_parameters.filename_));

    // If there is no such name, and this is NOT a fallback, add to the
    // vector
    if (it == print_parameters_vector.end())
    {
      if(!print_parameters.fallback_)
      {
        print_parameters_vector.push_back(print_parameters);
      }
    }
    else 
    {
      // we have found a prior parameter list with the same file name.
      // If the format matches, add this list to that pre-existing list.
      // If there is a format mismatch in a non-fallback, it is an error.
      // if the format mismatch is in a fallback, ignore both the error
      // and the attempted addition to the list.
      if ((*it).format_ != print_parameters.format_)
      {
        if (!(*it).fallback_)
        {
          Report::UserError() << "File " << print_parameters.filename_
                              << ": multiple print statements to this file with differing formats.";
        }
      }
      else
      {
        // Fly in the ointment:  some variable lists, such as those for
        // AC or HB_*, have had "Index" and "TIME" or "FREQUENCY" pushed into
        // the front.  We do *NOT* want to add these.
        Util::ParamList::const_iterator it2=print_parameters.variableList_.begin();
        while ( (*it2).tag() == "INDEX" ||
                (*it2).tag() == "TIME" ||
                (*it2).tag() == "FREQUENCY" )
        {
          ++it2;
        }
        (*it).variableList_.insert((*it).variableList_.end(),
                                 it2,
                                 print_parameters.variableList_.end());
      }
    }
  }
}


//-----------------------------------------------------------------------------
// Function      : OutputMgr::parsePRINTBlock
// Purpose       : registers set of variables to print for .PRINT
//
// Special Notes : ERK.  This function used to also output the header.
//                 Now, that functionality is down in outputPRINTHeader_.
//
//                 Also, this function allocates expression handlers for
//                 any expressions specified.
//
// Scope         : public
// Creator       : Robert Hoekstra, SNL, Parallel Computational Sciences
// Creation Date : 10/10/00
//-----------------------------------------------------------------------------
bool OutputMgr::parsePRINTBlock(const Util::OptionBlock & print_block)
{
  PrintParameters       print_parameters = defaultPrintParameters_;
  PrintType::PrintType  print_type;
  Format::Format        format = Format::STD;
  bool                  no_index = false;

  print_parameters.netlistLocation_ = print_block.getNetlistLocation();

  Util::ParamList::const_iterator iterParam = print_block.begin();
  for (; iterParam != print_block.end(); ++iterParam)
  {
    if (DEBUG_IO)
      Xyce::dout() << "iterParam->tag = " << iterParam->tag() << std::endl;

    if (iterParam->tag() == "WIDTH")
    {
      print_parameters.streamWidth_ = iterParam->getImmutableValue<int>();
    }
    else if (iterParam->tag() == "TYPE")
    {
      std::string s = iterParam->stringValue();
      if (s == "AC")
        print_type = PrintType::AC;
      else if (s == "AC_IC")
        print_type = PrintType::AC_IC;
      else if (s == "DC")
        print_type = PrintType::DC;
      else if (s == "HB")
        print_type = PrintType::HB;
      else if (s == "HB_TD")
        print_type = PrintType::HB_TD;
      else if (s == "HB_FD")
        print_type = PrintType::HB_FD;
      else if (s == "HB_IC")
        print_type = PrintType::HB_IC;
      else if (s == "HB_STARTUP")
        print_type = PrintType::HB_STARTUP;
      else if (s == "HOMOTOPY")
        print_type = PrintType::HOMOTOPY;
      else if (s == "MPDE")
        print_type = PrintType::MPDE;
      else if (s == "MPDE_IC")
        print_type = PrintType::MPDE_IC;
      else if (s == "MPDE_STARTUP")
        print_type = PrintType::MPDE_STARTUP;
      else if (s == "SENS")
        print_type = PrintType::SENS;
      else if (s == "TRANADJOINT")
        print_type = PrintType::TRANADJOINT;
      else if (s == "TRAN")
        print_type = PrintType::TRAN;
      else if (s == "NOISE")
        print_type = PrintType::NOISE;
      else if (s == "ROL")
        print_type = PrintType::DC; // TT
      else
      {
        Report::DevelFatal0() << "Unrecognized analysis type " << s;
      }
    }
    else if (iterParam->tag() == "PRECISION")
    {
      print_parameters.streamPrecision_ = iterParam->getImmutableValue<int>();
    }
    else if (iterParam->tag() == "FILTER")
    {
      print_parameters.filter_ = iterParam->getImmutableValue<double>();
    }
    else if (iterParam->tag() == "FORMAT")
    {
      std::string s = iterParam->stringValue();
      if (s == "STD")
        format = Format::STD;
      else if (s == "TECPLOT")
        format = Format::TECPLOT;
      else if (s == "DAKOTA")
        format = Format::DAKOTA;
      else if (s == "PROBE")
        format = Format::PROBE;
      else if (s == "NOINDEX")
      {
        format = Format::STD;
        no_index = true;
      }
      else if (s == "CSV")
      {
        format = Format::CSV;
        print_parameters.delimiter_ = ",";
      }
      else if (s == "RAW")
      {
        format = defaultPrintParameters_.asciiRaw_ ? Format::RAW_ASCII : Format::RAW;
      }
      else
      {
        Report::DevelFatal0() << "Unrecognized print format " << s;
      }
      print_parameters.format_ = format;
    }
    else if (iterParam->tag() == "TIMEWIDTH")
    {
      print_parameters.timeWidth_ = iterParam->getImmutableValue<int>();
    }
    else if (iterParam->tag() == "TIMESCALEFACTOR")
    {
      print_parameters.outputTimeScaleFactor_ = iterParam->getImmutableValue<double>();
    }
    else if (iterParam->tag() == "FILE")
    {
      // netlistFilename_ should be the default unless FILE was set to
      // something other than "" which is its default value.
      if (iterParam->stringValue() != "")
      {
        print_parameters.filename_ = iterParam->stringValue();
      }
    }
    else if (iterParam->tag() == "DELIMITER")
    {
      if (iterParam->stringValue() == "TAB")
      {
        print_parameters.delimiter_ = "\t";
      }
      else if (iterParam->stringValue() == "COMMA")
      {
        print_parameters.delimiter_ = ",";
      }
      else if (iterParam->stringValue() != "")
      {
        Report::UserWarning0() << "Invalid value of DELIMITER in .PRINT statment, ignoring";
      }
    }
    else
    {
      // This must be the first print variable.
      break;
    }
  }

  // Remaining parameters are variables
  print_parameters.variableList_.assign(iterParam, print_block.end());

  // Increase width to make sure that the columns do not run together
  if (print_parameters.streamWidth_ - print_parameters.streamPrecision_  < 9)
    print_parameters.streamWidth_ = print_parameters.streamPrecision_ + 9;

  // Indicate if an index column should be added
  print_parameters.printIndexColumn_ = !no_index
                                       && print_parameters.format_ != Format::CSV
                                       && print_parameters.format_ != Format::PROBE
                                       && print_parameters.format_ != Format::TECPLOT
                                       && print_parameters.format_ != Format::DAKOTA
                                       && print_parameters.format_ != Format::RAW
                                       && print_parameters.format_ != Format::RAW_ASCII;

  // Assemble the apropriate flavors of output variable lists based on the PRINT type and format
  if (print_type == PrintType::AC)
  {
    PrintParameters freq_print_parameters = print_parameters;
    if (freq_print_parameters.format_ != Format::PROBE)
    {
      freq_print_parameters.variableList_.push_front(Util::Param("FREQUENCY", 0.0));
    }
    if (freq_print_parameters.printIndexColumn_)
    {
      freq_print_parameters.variableList_.push_front(Util::Param("INDEX", 0.0));
    }
    freq_print_parameters.expandComplexTypes_ = freq_print_parameters.format_ != Format::PROBE
                                                && freq_print_parameters.format_ != Format::RAW
                                                && freq_print_parameters.format_ != Format::RAW_ASCII;
    addOutputPrintParameters(OutputType::AC, freq_print_parameters);

    PrintParameters ac_ic_print_parameters = print_parameters;
    ac_ic_print_parameters.fallback_ = true;
    ac_ic_print_parameters.filename_ = "";
    if (ac_ic_print_parameters.format_ != Format::PROBE)
    {
      ac_ic_print_parameters.variableList_.push_front(Util::Param("TIME", 0.0));
    }
    if (ac_ic_print_parameters.printIndexColumn_)
    {
      ac_ic_print_parameters.variableList_.push_front(Util::Param("INDEX", 0.0));
    }
    addOutputPrintParameters(OutputType::AC_IC, ac_ic_print_parameters);
  }
  else if (print_type == PrintType::AC_IC)
  {
    PrintParameters ac_ic_print_parameters = print_parameters;
    if (ac_ic_print_parameters.format_ != Format::PROBE)
    {
      ac_ic_print_parameters.variableList_.push_front(Util::Param("TIME", 0.0));
    }
    if (ac_ic_print_parameters.printIndexColumn_)
    {
      ac_ic_print_parameters.variableList_.push_front(Util::Param("INDEX", 0.0));
    }
    addOutputPrintParameters(OutputType::AC_IC, ac_ic_print_parameters);
  }
  else if (print_type == PrintType::NOISE)
  {
    PrintParameters noise_print_parameters = print_parameters;
    if (noise_print_parameters.format_ != Format::PROBE)
    {
      noise_print_parameters.variableList_.push_front(Util::Param("FREQUENCY", 0.0));
    }
    if (noise_print_parameters.printIndexColumn_)
    {
      noise_print_parameters.variableList_.push_front(Util::Param("INDEX", 0.0));
    }
    //noise_print_parameters.expandComplexTypes_ = noise_print_parameters.format_ != Format::PROBE
                                                //&& noise_print_parameters.format_ != Format::RAW
                                                //&& noise_print_parameters.format_ != Format::RAW_ASCII;

    if (noise_print_parameters.format_ == Format::STD)
    {
      noise_print_parameters.defaultExtension_ = ".NOISE.prn";
    }
    else if (noise_print_parameters.format_ == Format::TECPLOT)
    {
      noise_print_parameters.defaultExtension_ = ".NOISE.dat";
    }
    //else if (noise_print_parameters.format_ == Format::PROBE)
    //{
    //  noise_print_parameters.defaultExtension_ = ".NOISE.csd";
    //}
    //  else if (noise_print_parameters.format_ == Format::DAKOTA)
    //{
    //  noise_print_parameters.defaultExtension_ = ".NOISE.txt";
    //}
    else
    {
      noise_print_parameters.defaultExtension_ = ".NOISE_unknown";
    }

    std::copy(noiseVariableList_.begin(), noiseVariableList_.end(), std::back_inserter(noise_print_parameters.variableList_));

    addOutputPrintParameters(OutputType::NOISE, noise_print_parameters);

  }
  else if (print_type == PrintType::HOMOTOPY)
  {
    PrintParameters homotopy_print_parameters = print_parameters;
    if (homotopy_print_parameters.format_ == Format::STD)
    {
      homotopy_print_parameters.defaultExtension_ = ".HOMOTOPY.prn";
    }
    else if (homotopy_print_parameters.format_ == Format::CSV)
    {
      homotopy_print_parameters.defaultExtension_ = ".HOMOTOPY.prn";
      homotopy_print_parameters.format_ = Format::STD;
      homotopy_print_parameters.delimiter_ = "";
      
    }
    else if (homotopy_print_parameters.format_ == Format::TECPLOT)
    {
      homotopy_print_parameters.defaultExtension_ = ".HOMOTOPY.dat";
    }
    else if (homotopy_print_parameters.format_ == Format::PROBE)
    {
      homotopy_print_parameters.defaultExtension_ = ".HOMOTOPY.csd";
    }
    else
    {
      homotopy_print_parameters.defaultExtension_ = ".HOMOTOPY.unknown";
    }
    addOutputPrintParameters(OutputType::HOMOTOPY, homotopy_print_parameters);
  }
  else if (print_type == PrintType::SENS)
  {
    PrintParameters sensitivity_print_parameters = print_parameters;

    if (sensitivity_print_parameters.format_ == Format::STD)
    {
      sensitivity_print_parameters.defaultExtension_ = ".SENS.prn";
    }
    else if (sensitivity_print_parameters.format_ == Format::TECPLOT)
    {
      sensitivity_print_parameters.defaultExtension_ = ".SENS.dat";
    }
    else if (sensitivity_print_parameters.format_ == Format::PROBE)
    {
      sensitivity_print_parameters.defaultExtension_ = ".SENS.csd";
    }
    else if (sensitivity_print_parameters.format_ == Format::DAKOTA)
    {
      sensitivity_print_parameters.defaultExtension_ = ".SENS.txt";
    }
    else
    {
      sensitivity_print_parameters.defaultExtension_ = ".SENS.unknown";
    }
    std::copy(sensitivityVariableList_.begin(), sensitivityVariableList_.end(), std::back_inserter(sensitivity_print_parameters.variableList_));

    addOutputPrintParameters(OutputType::SENS, sensitivity_print_parameters);
  }
  else if (print_type == PrintType::TRANADJOINT)
  {
    PrintParameters transientAdjoint_print_parameters = print_parameters;

    if (transientAdjoint_print_parameters.format_ == Format::STD)
    {
      transientAdjoint_print_parameters.defaultExtension_ = ".TRADJ.prn";
    }
    else if (transientAdjoint_print_parameters.format_ == Format::TECPLOT)
    {
      transientAdjoint_print_parameters.defaultExtension_ = ".TRADJ.dat";
    }
    else
    {
      transientAdjoint_print_parameters.defaultExtension_ = ".TRADJ.unknown";
    }      
    // ERK: for now just using the same variable list, but change later.
    std::copy(
        sensitivityVariableList_.begin(), 
        sensitivityVariableList_.end(), 
        std::back_inserter(transientAdjoint_print_parameters.variableList_));

    addOutputPrintParameters(OutputType::TRANADJOINT, transientAdjoint_print_parameters);
  }
  else if (print_type == PrintType::HB)
  {
    PrintParameters freq_print_parameters = print_parameters;
    freq_print_parameters.fallback_ = true;
    if (freq_print_parameters.format_ != Format::PROBE)
      freq_print_parameters.variableList_.push_front(Util::Param("FREQUENCY", 0.0));
    if (freq_print_parameters.printIndexColumn_)
      freq_print_parameters.variableList_.push_front(Util::Param("INDEX", 0.0));
    freq_print_parameters.expandComplexTypes_ = freq_print_parameters.format_ != Format::PROBE
                                                && freq_print_parameters.format_ != Format::RAW
                                                && freq_print_parameters.format_ != Format::RAW_ASCII;
    addOutputPrintParameters(OutputType::HB_FD, freq_print_parameters);

    PrintParameters time_print_parameters = print_parameters;
    time_print_parameters.fallback_ = true;
    time_print_parameters.filename_ = "";
    if (time_print_parameters.format_ != Format::PROBE)
      time_print_parameters.variableList_.push_front(Util::Param("TIME", 0.0));
    if (time_print_parameters.printIndexColumn_)
      time_print_parameters.variableList_.push_front(Util::Param("INDEX", 0.0));
    addOutputPrintParameters(OutputType::HB_TD, time_print_parameters);

    PrintParameters hb_ic_print_parameters = print_parameters;
    hb_ic_print_parameters.fallback_ = true;
    hb_ic_print_parameters.filename_ = "";
    if (hb_ic_print_parameters.format_ == Format::STD)
      hb_ic_print_parameters.defaultExtension_ = ".hb_ic.prn";
    else if (hb_ic_print_parameters.format_ == Format::CSV)
      hb_ic_print_parameters.defaultExtension_ = ".hb_ic.csv";
    else if (hb_ic_print_parameters.format_ == Format::TECPLOT)
      hb_ic_print_parameters.defaultExtension_ = ".hb_ic.dat";
    else
      hb_ic_print_parameters.defaultExtension_ = ".hb_ic.unknown";
    
    if (hb_ic_print_parameters.format_ != Format::PROBE)
      hb_ic_print_parameters.variableList_.push_front(Util::Param("TIME", 0.0));
    if (hb_ic_print_parameters.printIndexColumn_)
      hb_ic_print_parameters.variableList_.push_front(Util::Param("INDEX", 0.0));
    addOutputPrintParameters(OutputType::HB_IC, hb_ic_print_parameters);

    PrintParameters hb_startup_print_parameters = print_parameters;
    hb_startup_print_parameters.fallback_ = true;
    hb_startup_print_parameters.filename_ = "";
    if (hb_startup_print_parameters.format_ == Format::STD)
      hb_startup_print_parameters.defaultExtension_ = ".startup.prn";
    else if (hb_startup_print_parameters.format_ == Format::CSV)
      hb_startup_print_parameters.defaultExtension_ = ".startup.csv";
    else if (hb_startup_print_parameters.format_ == Format::TECPLOT)
      hb_startup_print_parameters.defaultExtension_ = ".startup.dat";
    else
      hb_startup_print_parameters.defaultExtension_ = ".startup.unknown";
      
    if (hb_startup_print_parameters.format_ != Format::PROBE)
      hb_startup_print_parameters.variableList_.push_front(Util::Param("TIME", 0.0));
    if (hb_startup_print_parameters.printIndexColumn_)
      hb_startup_print_parameters.variableList_.push_front(Util::Param("INDEX", 0.0));
    addOutputPrintParameters(OutputType::HB_STARTUP, hb_startup_print_parameters);
  }
  else if (print_type == PrintType::HB_TD)
  {
    PrintParameters time_print_parameters = print_parameters;
    if (time_print_parameters.format_ != Format::PROBE)
      time_print_parameters.variableList_.push_front(Util::Param("TIME", 0.0));
    if (time_print_parameters.printIndexColumn_)
      time_print_parameters.variableList_.push_front(Util::Param("INDEX", 0.0));
    addOutputPrintParameters(OutputType::HB_TD, time_print_parameters);

    PrintParameters hb_ic_print_parameters = print_parameters;
    hb_ic_print_parameters.fallback_ = true;
    hb_ic_print_parameters.filename_ = "";
    if (hb_ic_print_parameters.format_ == Format::STD)
      hb_ic_print_parameters.defaultExtension_ = ".hb_ic.prn";
    else if (hb_ic_print_parameters.format_ == Format::CSV)
      hb_ic_print_parameters.defaultExtension_ = ".hb_ic.csv";
    else if (hb_ic_print_parameters.format_ == Format::TECPLOT)
      hb_ic_print_parameters.defaultExtension_ = ".hb_ic.dat";
    else
      hb_ic_print_parameters.defaultExtension_ = ".hb_ic.unknown";
      
    if (hb_ic_print_parameters.format_ != Format::PROBE)
      hb_ic_print_parameters.variableList_.push_front(Util::Param("TIME", 0.0));
    if (hb_ic_print_parameters.printIndexColumn_)
      hb_ic_print_parameters.variableList_.push_front(Util::Param("INDEX", 0.0));
    addOutputPrintParameters(OutputType::HB_IC, hb_ic_print_parameters);

    PrintParameters hb_startup_print_parameters = print_parameters;
    hb_startup_print_parameters.fallback_ = true;
    hb_startup_print_parameters.filename_ = "";
    if (hb_startup_print_parameters.format_ == Format::STD)
      hb_startup_print_parameters.defaultExtension_ = ".startup.prn";
    else if (hb_startup_print_parameters.format_ == Format::CSV)
      hb_startup_print_parameters.defaultExtension_ = ".startup.csv";
    else if (hb_startup_print_parameters.format_ == Format::TECPLOT)
      hb_startup_print_parameters.defaultExtension_ = ".startup.dat";
    else
      hb_startup_print_parameters.defaultExtension_ = ".startup.unknown";
      
    if (hb_startup_print_parameters.format_ != Format::PROBE)
      hb_startup_print_parameters.variableList_.push_front(Util::Param("TIME", 0.0));
    if (hb_startup_print_parameters.printIndexColumn_)
      hb_startup_print_parameters.variableList_.push_front(Util::Param("INDEX", 0.0));
    addOutputPrintParameters(OutputType::HB_STARTUP, hb_startup_print_parameters);
  }
  else if (print_type == PrintType::HB_FD)
  {
    PrintParameters freq_print_parameters = print_parameters;
    if (freq_print_parameters.format_ != Format::PROBE)
      freq_print_parameters.variableList_.push_front(Util::Param("FREQUENCY", 0.0));
    if (freq_print_parameters.printIndexColumn_)
      freq_print_parameters.variableList_.push_front(Util::Param("INDEX", 0.0));
    freq_print_parameters.expandComplexTypes_ = freq_print_parameters.format_ != Format::PROBE
                                                && freq_print_parameters.format_ != Format::RAW
                                                && freq_print_parameters.format_ != Format::RAW_ASCII;
    addOutputPrintParameters(OutputType::HB_FD, freq_print_parameters);

    PrintParameters hb_ic_print_parameters = print_parameters;
    hb_ic_print_parameters.fallback_ = true;
    hb_ic_print_parameters.filename_ = "";
    if (hb_ic_print_parameters.format_ == Format::STD)
      hb_ic_print_parameters.defaultExtension_ = ".hb_ic.prn";
    else if (hb_ic_print_parameters.format_ == Format::CSV)
      hb_ic_print_parameters.defaultExtension_ = ".hb_ic.csv";
    else if (hb_ic_print_parameters.format_ == Format::TECPLOT)
      hb_ic_print_parameters.defaultExtension_ = ".hb_ic.dat";
    else
      hb_ic_print_parameters.defaultExtension_ = ".hb_ic.unknown";
    
    if (hb_ic_print_parameters.format_ != Format::PROBE)
      hb_ic_print_parameters.variableList_.push_front(Util::Param("TIME", 0.0));
    if (hb_ic_print_parameters.printIndexColumn_)
      hb_ic_print_parameters.variableList_.push_front(Util::Param("INDEX", 0.0));
    addOutputPrintParameters(OutputType::HB_IC, hb_ic_print_parameters);

    PrintParameters hb_startup_print_parameters = print_parameters;
    hb_startup_print_parameters.fallback_ = true;
    hb_startup_print_parameters.filename_ = "";
    if (hb_startup_print_parameters.format_ == Format::STD)
      hb_startup_print_parameters.defaultExtension_ = ".startup.prn";
    else if (hb_startup_print_parameters.format_ == Format::CSV)
      hb_startup_print_parameters.defaultExtension_ = ".startup.csv";
    else if (hb_startup_print_parameters.format_ == Format::TECPLOT)
      hb_startup_print_parameters.defaultExtension_ = ".startup.dat";
    else
      hb_startup_print_parameters.defaultExtension_ = ".startup.unknown";
    
    if (hb_startup_print_parameters.format_ != Format::PROBE)
      hb_startup_print_parameters.variableList_.push_front(Util::Param("TIME", 0.0));
    if (hb_startup_print_parameters.printIndexColumn_)
      hb_startup_print_parameters.variableList_.push_front(Util::Param("INDEX", 0.0));
    addOutputPrintParameters(OutputType::HB_STARTUP, hb_startup_print_parameters);
  }
  else if (print_type == PrintType::HB_IC)
  {
    PrintParameters hb_ic_print_parameters = print_parameters;
    if (hb_ic_print_parameters.format_ == Format::STD)
      hb_ic_print_parameters.defaultExtension_ = ".hb_ic.prn";
    else if (hb_ic_print_parameters.format_ == Format::CSV)
      hb_ic_print_parameters.defaultExtension_ = ".hb_ic.csv";
    else if (hb_ic_print_parameters.format_ == Format::TECPLOT)
      hb_ic_print_parameters.defaultExtension_ = ".hb_ic.dat";
    else
      hb_ic_print_parameters.defaultExtension_ = ".hb_ic.unknown";
      
    if (hb_ic_print_parameters.format_ != Format::PROBE)
      hb_ic_print_parameters.variableList_.push_front(Util::Param("TIME", 0.0));
    if (hb_ic_print_parameters.printIndexColumn_)
      hb_ic_print_parameters.variableList_.push_front(Util::Param("INDEX", 0.0));
    addOutputPrintParameters(OutputType::HB_IC, hb_ic_print_parameters);
  }
  else if (print_type == PrintType::HB_STARTUP)
  {
    PrintParameters hb_startup_print_parameters = print_parameters;
    if (hb_startup_print_parameters.format_ == Format::STD)
      hb_startup_print_parameters.defaultExtension_ = ".startup.prn";
    else if (hb_startup_print_parameters.format_ == Format::CSV)
      hb_startup_print_parameters.defaultExtension_ = ".startup.csv";
    else if (hb_startup_print_parameters.format_ == Format::TECPLOT)
      hb_startup_print_parameters.defaultExtension_ = ".startup.dat";
    else
      hb_startup_print_parameters.defaultExtension_ = ".startup.unknown";
      
    if (hb_startup_print_parameters.format_ != Format::PROBE)
      hb_startup_print_parameters.variableList_.push_front(Util::Param("TIME", 0.0));
    if (hb_startup_print_parameters.printIndexColumn_)
      hb_startup_print_parameters.variableList_.push_front(Util::Param("INDEX", 0.0));
    addOutputPrintParameters(OutputType::HB_STARTUP, hb_startup_print_parameters);
  }
  else if (print_type == PrintType::MPDE)
  {
    PrintParameters mpde_print_parameters = print_parameters;
    addOutputPrintParameters(OutputType::MPDE, mpde_print_parameters);

    PrintParameters mpde_ic_print_parameters = print_parameters;
    mpde_ic_print_parameters.fallback_ = true;
    mpde_ic_print_parameters.defaultExtension_ = ".mpde_ic.prn";

    if (mpde_ic_print_parameters.format_ != Format::PROBE)
      mpde_ic_print_parameters.variableList_.push_front(Util::Param("TIME", 0.0));
    if (mpde_ic_print_parameters.printIndexColumn_)
      mpde_ic_print_parameters.variableList_.push_front(Util::Param("INDEX", 0.0));
    addOutputPrintParameters(OutputType::MPDE_IC, mpde_ic_print_parameters);

    PrintParameters mpde_startup_print_parameters = print_parameters;
    mpde_startup_print_parameters.fallback_ = true;
    mpde_startup_print_parameters.filename_ = "";
    if (mpde_startup_print_parameters.format_ == Format::STD)
      mpde_startup_print_parameters.defaultExtension_ = ".startup.prn";
    else if (mpde_startup_print_parameters.format_ == Format::CSV)
      mpde_startup_print_parameters.defaultExtension_ = ".startup.csv";
    else if (mpde_startup_print_parameters.format_ == Format::TECPLOT)
      mpde_startup_print_parameters.defaultExtension_ = ".startup.dat";
    else
      mpde_startup_print_parameters.defaultExtension_ = ".startup.unknown";
      
    if (mpde_startup_print_parameters.format_ != Format::PROBE)
      mpde_startup_print_parameters.variableList_.push_front(Util::Param("TIME", 0.0));
    if (mpde_startup_print_parameters.printIndexColumn_)
      mpde_startup_print_parameters.variableList_.push_front(Util::Param("INDEX", 0.0));
    addOutputPrintParameters(OutputType::MPDE_STARTUP, mpde_startup_print_parameters);
  }
  else if (print_type == PrintType::MPDE_IC)
  {
    PrintParameters mpde_ic_print_parameters = print_parameters;
    mpde_ic_print_parameters.defaultExtension_ = ".mpde_ic.prn";

    if (mpde_ic_print_parameters.format_ != Format::PROBE)
      mpde_ic_print_parameters.variableList_.push_front(Util::Param("TIME", 0.0));
    if (mpde_ic_print_parameters.printIndexColumn_)
      mpde_ic_print_parameters.variableList_.push_front(Util::Param("INDEX", 0.0));
    addOutputPrintParameters(OutputType::MPDE_IC, mpde_ic_print_parameters);
  }
  else if (print_type == PrintType::MPDE_STARTUP)
  {
    PrintParameters mpde_startup_print_parameters = print_parameters;
    if (mpde_startup_print_parameters.format_ == Format::STD)
      mpde_startup_print_parameters.defaultExtension_ = ".startup.prn";
    else if (mpde_startup_print_parameters.format_ == Format::CSV)
      mpde_startup_print_parameters.defaultExtension_ = ".startup.csv";
    else if (mpde_startup_print_parameters.format_ == Format::TECPLOT)
      mpde_startup_print_parameters.defaultExtension_ = ".startup.dat";
    else
      mpde_startup_print_parameters.defaultExtension_ = ".startup.unknown";
      
    if (mpde_startup_print_parameters.format_ != Format::PROBE)
      mpde_startup_print_parameters.variableList_.push_front(Util::Param("TIME", 0.0));
    if (mpde_startup_print_parameters.printIndexColumn_)
      mpde_startup_print_parameters.variableList_.push_front(Util::Param("INDEX", 0.0));
    addOutputPrintParameters(OutputType::MPDE_STARTUP, mpde_startup_print_parameters);
  }
  else if (print_type == PrintType::TRAN)
  {
    addOutputPrintParameters(OutputType::TRAN, print_parameters);
    PrintParameters homotopy_print_parameters = print_parameters;
    homotopy_print_parameters.fallback_ = true;
    homotopy_print_parameters.filename_ = "";
    if (homotopy_print_parameters.format_ == Format::STD)
    {
      homotopy_print_parameters.defaultExtension_ = ".HOMOTOPY.prn";
    }
    else if (homotopy_print_parameters.format_ == Format::CSV)
    {
      // There is no homotopy output formatter for .csv
      homotopy_print_parameters.defaultExtension_ = ".HOMOTOPY.prn";
      homotopy_print_parameters.format_ = Format::STD;
      homotopy_print_parameters.delimiter_ = "";
    }
    else if (homotopy_print_parameters.format_ == Format::TECPLOT)
    {
      homotopy_print_parameters.defaultExtension_ = ".HOMOTOPY.dat";
    }
    else if (homotopy_print_parameters.format_ == Format::PROBE)
    {
      homotopy_print_parameters.defaultExtension_ = ".HOMOTOPY.csd";
    }
    else
    {
      homotopy_print_parameters.defaultExtension_ = ".HOMOTOPY.unknown";
    }
      
    addOutputPrintParameters(OutputType::HOMOTOPY, homotopy_print_parameters);

    PrintParameters mpde_print_parameters = print_parameters;
    mpde_print_parameters.fallback_ = true;
    mpde_print_parameters.filename_ = "";
    addOutputPrintParameters(OutputType::MPDE, mpde_print_parameters);

    PrintParameters mpde_ic_print_parameters = print_parameters;
    mpde_ic_print_parameters.fallback_ = true;
    mpde_ic_print_parameters.filename_ = "";
    mpde_ic_print_parameters.defaultExtension_ = ".mpde_ic.prn";

    if (mpde_ic_print_parameters.format_ != Format::PROBE)
      mpde_ic_print_parameters.variableList_.push_front(Util::Param("TIME", 0.0));
    if (mpde_ic_print_parameters.printIndexColumn_)
      mpde_ic_print_parameters.variableList_.push_front(Util::Param("INDEX", 0.0));
    addOutputPrintParameters(OutputType::MPDE_IC, mpde_ic_print_parameters);
  }
  else if (print_type == PrintType::DC)
  {
    PrintParameters dc_print_parameters = print_parameters;
    if (dc_print_parameters.format_ == Format::STD)
      dc_print_parameters.defaultExtension_ = ".prn";
    addOutputPrintParameters(OutputType::DC, dc_print_parameters);
  }
  else
  {
    Report::UserError0() << "Unrecognized .PRINT type";
  }

  activeOutputterStack_.push_back(std::vector<Outputter::Interface *>());

  return true;
}

//-----------------------------------------------------------------------------
// Function      : OutputMgr::registerSens
// Purpose       : registers set of variables to set for .SENS.
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter, SNL, Electrical and Microsystems Modeling
// Creation Date : 2/10/2014
//-----------------------------------------------------------------------------
bool OutputMgr::registerSens(const Util::OptionBlock &option_block)
{
  bool bsuccess = true;

  std::vector<std::string> functions;
  std::vector<std::string> parameters;

  for (Util::ParamList::const_iterator it = option_block.begin(), end = option_block.end(); it != end; ++it)
  {
    std::string tag = (*it).uTag();
    if ( std::string( (*it).uTag() ,0,7) == "OBJFUNC") // this is a vector
    {
      functions.push_back((*it).stringValue());
    }
    else if (std::string((*it).uTag(), 0, 5) == "PARAM")
    {
      parameters.push_back((*it).stringValue());
    }
    else
    {
      Xyce::Report::UserWarning() << (*it).uTag() << " is not a recognized sensitivity solver option.\n" << std::endl;
    }
  }

  // ERK: important to set the index properly to make multiple objectives work correctly.
  //
  // For sensitivity outputs, a bunch of new variables are automatically added to the variable
  // list.  For now, they pretend to have the keyword "SENS", as this is an acceptable keyword
  // in the parser. (I tried using something like "TMPSENS" and the parser barfed.  In addition
  // to the keyword "SENS", they also have the names of the parameters included to later flesh out
  // the sensitivity names.
  //
  // Later, these lists of variables will be properly set up in the N_IO_OpBuilders.C file, which is
  // called from the OutputMgr::checkPrintParameters function.
  int index = 0;
  int index2 = 0;
  for (std::vector<std::string>::const_iterator it1 = functions.begin(), end1 = functions.end(); it1 != end1; ++it1,++index2)
  {
    const std::string &function = (*it1);

    {
      Util::Marshal mout;
      mout << function << std::string("OBJECTIVEFUNCTION") << Util::Op::identifier<SensitivityObjFunctionOp>() << index2;
      sensitivityVariableList_.push_back(Util::Param("SENS", mout.str()));
    }

    for (std::vector<std::string>::const_iterator it2 = parameters.begin(), end2 = parameters.end(); it2 != end2; ++it2, ++index)
    {
      const std::string &parameter = (*it2);

      if (sensitivityOptions_ & SensitivityOptions::DIRECT)
      {
        if (sensitivityOptions_ & SensitivityOptions::UNSCALED)
        {
          Util::Marshal mout;
          mout << function << parameter << Util::Op::identifier<SensitivitydOdpDirectOp>() << index;
          sensitivityVariableList_.push_back(Util::Param("SENS", mout.str()));
        }
        if (sensitivityOptions_ & SensitivityOptions::SCALED)
        {
          Util::Marshal mout;
          mout << function << parameter << Util::Op::identifier<SensitivitydOdpDirectScaledOp>() << index;
          sensitivityVariableList_.push_back(Util::Param("SENS", mout.str()));
        }
      }

      if (sensitivityOptions_ & SensitivityOptions::ADJOINT)
      {
        if (sensitivityOptions_ & SensitivityOptions::UNSCALED)
        {
          Util::Marshal mout;
          mout << function << parameter << Util::Op::identifier<SensitivitydOdpAdjointOp>() << index;
          sensitivityVariableList_.push_back(Util::Param("SENS", mout.str()));
        }
        if (sensitivityOptions_ & SensitivityOptions::SCALED)
        {
          Util::Marshal mout;
          mout << function << parameter << Util::Op::identifier<SensitivitydOdpAdjointScaledOp>() << index;
          sensitivityVariableList_.push_back(Util::Param("SENS", mout.str()));
        }

        adjointSensitivityFlag_ = true;
      }
    }
  }

  enableSensitivityFlag_ = true;

  return bsuccess;
}

//-----------------------------------------------------------------------------
// Function      : OutputMgr::registerSensOptions
// Purpose       : registers set of variables to set for .OPTIONS SENSITIVITY
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter, SNL, Electrical and Microsystems Modeling
// Creation Date : 2/10/2014
//-----------------------------------------------------------------------------
bool OutputMgr::registerSensOptions(const Util::OptionBlock &option_block)
{
  sensitivityOptions_ = 0;

  bool adjointGiven=false;
  bool outputUnscaledGiven=false;

  sensitivityOptions_ =0;

  Util::ParamList::const_iterator it = option_block.begin(); 
  Util::ParamList::const_iterator end = option_block.end();
  for ( ; it != end; ++it)
  {
    if ((*it).uTag() == "ADJOINT")
    {
      adjointGiven=true;
      if ((*it).getImmutableValue<bool>())
      {
        sensitivityOptions_ |= SensitivityOptions::ADJOINT;
      }
    }
    else if ((*it).uTag() == "DIRECT" && (*it).getImmutableValue<bool>())
    {
      sensitivityOptions_ |= SensitivityOptions::DIRECT;
    }
    else if ((*it).uTag() == "OUTPUTSCALED" && (*it).getImmutableValue<bool>())
    {
      sensitivityOptions_ |= SensitivityOptions::SCALED;
    }
    else if ((*it).uTag() == "OUTPUTUNSCALED")
    {
      outputUnscaledGiven=true;
      if ((*it).getImmutableValue<bool>())
      {
        sensitivityOptions_ |= SensitivityOptions::UNSCALED;
      }
    }
  }

  // default behavior is for the code to assume adjoint sensitivites are wanted
  // even if ADJOINT=1 wasn't specified.
  if (!adjointGiven)
  {
    sensitivityOptions_ |= SensitivityOptions::ADJOINT;
  }

  // default behavior is for the code to assume unscaled sensitivities are wanted
  // for output.
  if (!outputUnscaledGiven)
  {
    sensitivityOptions_ |= SensitivityOptions::UNSCALED;
  }

  return true;
}

//-----------------------------------------------------------------------------
// Function      : OutputMgr::registerNoise
// Purpose       : registers set of variables to set for .NOISE.
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter, SNL
// Creation Date : 1/26/2015
//-----------------------------------------------------------------------------
bool OutputMgr::registerNoise (const Util::OptionBlock &option_block)
{
  bool bsuccess = true;

  std::vector<std::string> parameters;
  for (Util::ParamList::const_iterator it = option_block.begin(), 
      end = option_block.end(); it != end; ++it)
  { 
    // if pts_per_summary is set, then we need to output a lot more stuff.
    // NOTE.  pts_per_summary is a spice3-ism, and probably not that useful here.
    if ((*it).uTag() == "PTS_PER_SUMMARY")
    {
      pts_per_summary_ = (*it).getImmutableValue<int>();
      pts_per_summary_Given = true;
    }
  }

  return bsuccess;
}

//-----------------------------------------------------------------------------
// Function      : makeParamList
// Purpose       : 
// Special Notes : 
// Scope         :
// Creator       : 
// Creation Date : 
//-----------------------------------------------------------------------------
template<class It, class Out>
void makeParamList(const std::string name, It first, It last, Out out)
{
  for (; first != last; ++first)
  {
    *out++ = Util::Param(name, 1);
    *out++ = Util::Param(*first, 0.0);
  }
}

//-----------------------------------------------------------------------------
// Function      : Xyce::IO::removeStarVariables
// Purpose       : Process V(*) and I(*) on .print line
// Special Notes : Replaces v(*) and i(*) that appears on the .print line
//                 with a list of all v() or i() variables as appropriate.
// Scope         :
// Creator       : David Baur, Raytheon
// Creation Date : 6/28/2013
//-----------------------------------------------------------------------------
void removeStarVariables(
  Parallel::Machine     comm,
  Util::ParamList &     variable_list,
  const NodeNameMap &   all_nodes,
  const NodeNameMap &   external_nodes)
{
  bool vStarFound = false;
  bool iStarFound = false;

  Util::ParamList::iterator vStarPosition = variable_list.end();
  Util::ParamList::iterator iStarPosition = variable_list.end();

  // remove the v(*) and i(*) from the .print line
  Util::ParamList::iterator it = variable_list.begin();
  for ( ; it != variable_list.end(); )
  {
    // process the * entries
    if ((*it).tag() == "*")
    {
      // move to the type
      --it;

      // remember type of replacement and location of *
      if ((*it).tag() == "V")
      {
        vStarFound = true;
        vStarPosition = it;
        --vStarPosition;
      }
      else if ((*it).tag() == "I")
      {
        iStarFound = true;
        iStarPosition = it;
        --iStarPosition;
      }

      // remove the v or i
      it = variable_list.erase(it);

      // remove the *
      it = variable_list.erase(it);
    }

    // move to next item on .print line
    else
    {
      ++it;
    }
  }

  std::vector<std::string> v_list;
  std::vector<std::string> i_list;

  if (vStarFound)
  {
    NodeNameMap::const_iterator it = external_nodes.begin(); 
    for ( ; it != external_nodes.end() ; ++it)
    {
      ExtendedString tmpStr((*it).first);
      tmpStr.toUpper();

      if (tmpStr.rfind("BRANCH") == std::string::npos)
        v_list.push_back(tmpStr);
    }
    ++vStarPosition;
  }

  if (iStarFound)
  {
    NodeNameMap::const_iterator iter_a = all_nodes.begin();
    for ( ; iter_a != all_nodes.end() ; ++iter_a)
    {
      ExtendedString tmpStr((*iter_a).first);
      tmpStr.toUpper();

      size_t pos = tmpStr.rfind("BRANCH");
      if (pos != std::string::npos && tmpStr[0] != 'Y')
      {
        tmpStr = tmpStr.substr(0, pos - 1);

        i_list.push_back(tmpStr);
      }
    }
    ++iStarPosition;
  }

  Util::Marshal mout;
  mout << v_list << i_list;

  std::vector<std::string> dest;
  Parallel::AllGatherV(comm, mout.str(), dest);

  Util::ParamList vStarList;
  Util::ParamList iStarList;

  for (int p = 0; p < Parallel::size(comm); ++p)
  {
    Util::Marshal min(dest[p]);

    std::vector<std::string> x;
    std::vector<std::string> y;
    min >> x >> y;
    makeParamList("V", x.begin(), x.end(), std::back_inserter(vStarList));
    makeParamList("I", y.begin(), y.end(), std::back_inserter(iStarList));
  }

  // append temporary lists to print block, erasing temporary lists
  variable_list.splice(vStarPosition, vStarList);
  variable_list.splice(iStarPosition, iStarList);
}

//-----------------------------------------------------------------------------
// Function      : OutputMgr::setSweepParameters
// Purpose       : Copy .DC or .STEP sweep parameters, and set up flags if
//                 sweeping the TEMP variable.
// Special Notes :
// Scope         : public
// Creator       : David Baur, Raytheon
// Creation Date : 6/28/2013
//-----------------------------------------------------------------------------
void OutputMgr::setStepSweepVector(const Analysis::SweepVector &sweep_vector)
{
  if (!sweep_vector.empty())
  {
    outputState_.stepSweepVector_ =  sweep_vector;
  }

  // check if one of the sweep variables is temperature.
  //(this only needs to be checked one time.)
  if (!sweep_vector.empty())
  {
    for (Analysis::SweepVector::const_iterator iterP = outputState_.stepSweepVector_.begin(), 
        end = outputState_.stepSweepVector_.end(); iterP != end; ++iterP)
    {
      if (equal_nocase((*iterP).name, "TEMP"))
      {
        outputState_.tempSweepFlag_ = true;
      }
    }
  }
}

//-----------------------------------------------------------------------------
// Function      : OutputMgr::setSweepParameters
// Purpose       : Copy .DC or .STEP sweep parameters, and set up flags if
//                 sweeping the TEMP variable.
// Special Notes :
// Scope         : public
// Creator       : David Baur, Raytheon
// Creation Date : 6/28/2013
//-----------------------------------------------------------------------------
void OutputMgr::setDCSweepVector(const Analysis::SweepVector &sweep_vector)
{
  if (!sweep_vector.empty())
  {
    outputState_.dcSweepVector_ = sweep_vector;
  }

  // check if one of the DC sweep variables is temperature.
  //(this only needs to be checked one time.)
  if (!sweep_vector.empty() && !outputCalledBefore_)
  {
    Analysis::SweepVector::const_iterator iterP;
    for (iterP = outputState_.dcSweepVector_.begin();
         iterP != outputState_.dcSweepVector_.end(); ++iterP)
    {
      if (equal_nocase((*iterP).name, "TEMP"))
      {
        outputState_.tempSweepFlag_ = true;
      }
    }
  }
}

//-----------------------------------------------------------------------------
// Function      : OutputMgr::fixupPrintParameters
// Purpose       : Perform some .print line checks and munging, primarily
//                 dealing with use of V(*) and I(*)
// Special Notes : Mostly a re-wrapping of work that used to be done in the
//                 check_output function, which was too sprawling to remain
//                 as a single function.
// Scope         : public
// Creator       : David Baur, Raytheon
// Creation Date : 6/26/2013
//-----------------------------------------------------------------------------
void OutputMgr::fixupPrintParameters(
    Parallel::Machine comm, 
    PrintParameters &print_parameters)
{
  removeStarVariables(comm, print_parameters.variableList_, 
      getSolutionNodeMap(), getExternalNodeMap());
}

//-----------------------------------------------------------------------------
// Function      : OutputMgr::output
// Purpose       : Runs specified output commands
//
// Special Notes : ERK.  I've refactored this so that it receives STEP
//                 and DC sweep parameter information(mainly name and value).
//                 This function is called from the time integrator, which has
//                 all this info.
//
// Scope         : public
// Creator       : Robert Hoekstra, SNL, Parallel Computational Sciences
// Creation Date : 10/10/00
//-----------------------------------------------------------------------------
void OutputMgr::output(
  Parallel::Machine             comm,
  const double                  time,
  const double                  circuit_temp, 
  const int                     stepNumber,
  const int                     maxStep,
  const Analysis::SweepVector & step_sweep_vector,
  const int                     dcNumber,
  const int                     maxDC,
  const Analysis::SweepVector & dc_sweep_vector,
  const Linear::Vector &        solnVecPtr,
  const Linear::Vector &        stateVecPtr,
  const Linear::Vector &        storeVecPtr,
  const Linear::Vector &        lead_current_vector,
  const Linear::Vector &        junction_voltage_vector,
  const Linear::Vector &        lead_current_dqdt_vector,
  const std::vector<double> &   objectiveVec,
  const std::vector<double> &   dOdpVec,
  const std::vector<double> &   dOdpAdjVec,
  const std::vector<double> &   scaled_dOdpVec,
  const std::vector<double> &   scaled_dOdpAdjVec,
  bool                          skipPrintLineOutput)
{
  // copy over time:
  outputState_.circuitTime_ = time;

  // copy over the step sweep information:
  outputState_.stepLoopNumber_ = stepNumber;
  outputState_.stepMaxCount_ = maxStep;

  // copy the new values into the locally owned vector:
  if (!step_sweep_vector.empty())
  {
    outputState_.stepSweepVector_ = step_sweep_vector;
  }

  // copy over the dc sweep information:
  dcLoopNumber_ = dcNumber;
  maxDCSteps_ = maxDC;

  // if (!analysisManager_->getBlockAnalysisFlag())
  // {
  //   // This error test should not be used in the MPDE case, as at least
  //   // one of the initial conditions that can be
  //   // used by MPDE is a DC sweep.
  //   if (maxDCSteps_ > 0 && analysisManager_->getTransientFlag())
  //   {
  //     Report::DevelFatal0() << "Print type is inconsistent, maxDCSteps = " << maxDCSteps_;
  //   }
  // }

  if (!dc_sweep_vector.empty())
  {
    outputState_.dcSweepVector_ = dc_sweep_vector;

    // For now just have PRINTdcvalue, etc just be the first parameter.
    const Analysis::SweepParam &firstParam = dc_sweep_vector.front();
    PRINTdcname_   = firstParam.name;
    PRINTdcvalue_  = firstParam.currentVal;
    if (firstParam.type == "LIST")
    {
      PRINTdcstart_  = firstParam.valList.front();
      PRINTdcstop_   = firstParam.valList.back();
    }
    else
    {
      PRINTdcstart_  = firstParam.startVal;
      PRINTdcstop_   = firstParam.stopVal;
    }
  }

  // Check for temperature:
  outputState_.circuitTemp_ = circuit_temp;

  // Needs to pass skipPrintLineOutput
  if (!skipPrintLineOutput)
  {
    if (!activeOutputterStack_.empty())
    {
      for (std::vector<Outputter::Interface *>::const_iterator
          it = activeOutputterStack_.back().begin();
          it != activeOutputterStack_.back().end(); ++it)
      {
        (*it)->output(comm, solnVecPtr, stateVecPtr, storeVecPtr, 
                      lead_current_vector, junction_voltage_vector);

        (*it)->outputSensitivity(comm, objectiveVec,
                                 dOdpVec, dOdpAdjVec, 
                                 scaled_dOdpVec, scaled_dOdpAdjVec,
                                 solnVecPtr, stateVecPtr, storeVecPtr);
      }
    }
  }

  // This will become an outputter
  if (hdf5FileNameGiven_)
  {
    updateHDF5Output(comm, solnVecPtr);
  }
  
  outputCalledBefore_ = true;
}

//-----------------------------------------------------------------------------
// Function      : OutputMgr::outputSensitivity
// Purpose       : outputs sensitivity info  
//
// Special Notes : This was set up specifically to support transient adjoint 
//                 sensitivities.  Specifically for the case of using them to 
//                 compute local sensitivities.  It is (mostly) a copy of the
//                 OutputMgr::output function, but with the various 
//                 non-sensitivity stuff stripped out.
//
// Scope         : public
// Creator       : Eric Keiter, SNL
// Creation Date : 4/13/2016
//-----------------------------------------------------------------------------
void OutputMgr::outputSensitivity(
  Parallel::Machine             comm,
  const double                  time,
  const double                  circuit_temp, 
  const int                     stepNumber,
  const int                     maxStep,
  const Analysis::SweepVector & step_sweep_vector,
  const int                     dcNumber,
  const int                     maxDC,
  const Analysis::SweepVector & dc_sweep_vector,
  const Linear::Vector &        solnVecPtr,
  const Linear::Vector &        stateVecPtr,
  const Linear::Vector &        storeVecPtr,
  const Linear::Vector &        lead_current_vector,
  const Linear::Vector &        junction_voltage_vector,
  const Linear::Vector &        lead_current_dqdt_vector,
  const std::vector<double> &   objectiveVec,
  const std::vector<double> &   dOdpVec,
  const std::vector<double> &   dOdpAdjVec,
  const std::vector<double> &   scaled_dOdpVec,
  const std::vector<double> &   scaled_dOdpAdjVec,
  bool                          skipPrintLineOutput)
{
  // copy over time:
  outputState_.circuitTime_ = time;

  // copy over the step sweep information:
  outputState_.stepLoopNumber_ = stepNumber;
  outputState_.stepMaxCount_ = maxStep;

  // copy the new values into the locally owned vector:
  if (!step_sweep_vector.empty())
  {
    outputState_.stepSweepVector_ = step_sweep_vector;
  }

  // Check for temperature:
  outputState_.circuitTemp_ = circuit_temp;

  // Needs to pass skipPrintLineOutput
  if (!skipPrintLineOutput)
  {
    OutputterMap::iterator find_it = outputterMap_.find( PrintType::TRANADJOINT );

    if (find_it != outputterMap_.end())
    {
      std::vector<Outputter::Interface *>::const_iterator it = (*find_it).second.begin();
      std::vector<Outputter::Interface *>::const_iterator end = (*find_it).second.end();

      for ( ; it != end; ++it)
      {
        (*it)->outputSensitivity(comm, objectiveVec,
                                 dOdpVec, dOdpAdjVec, 
                                 scaled_dOdpVec, scaled_dOdpAdjVec,
                                 solnVecPtr, stateVecPtr, storeVecPtr);
      }
    }
    else
    {
      Report::UserWarning0() << "Cannot find any transient adjoint outputters!";
    }
  }
}


//-----------------------------------------------------------------------------
// Function      : OutputMgr::finishSensitivityOutput
// Purpose       : Runs specified finish output commands for adjoint 
//                 sensitivities.
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter
// Creation Date : 5/3/2016
//-----------------------------------------------------------------------------
void OutputMgr::finishSensitivityOutput()
{
  OutputterMap::iterator find_it = outputterMap_.find( PrintType::TRANADJOINT );
  if (find_it != outputterMap_.end())
  {
    std::vector<Outputter::Interface *>::const_iterator it = (*find_it).second.begin();
    std::vector<Outputter::Interface *>::const_iterator end = (*find_it).second.end();

    for ( ; it != end; ++it)
    {
      (*it)->finishOutput();
    }
  }
  else
  {
    //Report::UserWarning0() << "Cannot find any transient adjoint outputters!";
  }
}

//-----------------------------------------------------------------------------
// Function      : OutputMgr::outputAC
// Purpose       : .PRINT output for ac runs
// Special Notes :
// Scope         : public
// Creator       : Todd Coffey
// Creation Date : 8/5/08
//-----------------------------------------------------------------------------
void OutputMgr::outputAC(
  Parallel::Machine     comm,
  double                frequency,
  const Linear::Vector &  real_solution_vector,
  const Linear::Vector &  imaginary_solution_vector)
{
  outputState_.circuitFrequency_ = frequency;

  if (!activeOutputterStack_.empty())
  {
    std::vector<Outputter::Interface *>::const_iterator it = 
      activeOutputterStack_.back().begin(); 

    for ( ; it != activeOutputterStack_.back().end(); ++it)
    {
      (*it)->outputAC(comm, frequency, real_solution_vector, imaginary_solution_vector);
    }
  }
}


//-----------------------------------------------------------------------------
// Function      : OutputMgr::outputNoise
// Purpose       : .PRINT output for noise runs
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter
// Creation Date : 8/5/08
//-----------------------------------------------------------------------------
void OutputMgr::outputNoise(
  Parallel::Machine     comm,
  double                frequency,
    double              totalOutputNoiseDens_, 
    double              totalInputNoiseDens_, 
    const std::vector<Xyce::Analysis::NoiseData*> & noiseDataVec_
    )
{
  outputState_.circuitFrequency_ = frequency;

  if (!activeOutputterStack_.empty())
  {
    std::vector<Outputter::Interface *>::const_iterator it = 
      activeOutputterStack_.back().begin();

    for ( ; it != activeOutputterStack_.back().end(); ++it)
    {
      (*it)->outputNoise(comm, frequency, totalOutputNoiseDens_, 
          totalInputNoiseDens_, noiseDataVec_);
    }
  }
}

//-----------------------------------------------------------------------------
// Function      : OutputMgr::outputHomotopy
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Todd Coffey
// Creation Date : 8/5/08
//-----------------------------------------------------------------------------
void OutputMgr::outputHomotopy(
  Parallel::Machine                     comm,
  const std::vector<std::string> &      parameter_names,
  const std::vector<double> &           param_values,
  const Linear::Vector &                  solution_vector)
{
  if (!activeOutputterStack_.empty())
  {
    for (std::vector<Outputter::Interface *>::const_iterator
        it = activeOutputterStack_.back().begin();
        it != activeOutputterStack_.back().end(); ++it)
    {
      (*it)->outputHomotopy(comm, parameter_names, param_values, solution_vector);
    }
  }
}

//-----------------------------------------------------------------------------
// Function      : OutputMgr::finishOutput
// Purpose       : Runs specified finish output commands
// Special Notes :
// Scope         : public
// Creator       : Robert Hoekstra, SNL, Parallel Computational Sciences
// Creation Date : 12/13/00
//-----------------------------------------------------------------------------
void OutputMgr::finishOutput()
{
  if (!activeOutputterStack_.empty())
  {
    for (std::vector<Outputter::Interface *>::const_iterator
        it = activeOutputterStack_.back().begin();
        it != activeOutputterStack_.back().end(); ++it)
    {
      (*it)->finishOutput();
    }
  }

  if (hdf5FileNameGiven_)
  {
    closeHDF5Output();
  }
}

//-----------------------------------------------------------------------------
// Function      : OutputMgr::outputHB
// Purpose       : .print output for Harmonic Balance runs
// Special Notes :
// Scope         : public
// Creator       : David Baur, Raytheon
// Creation Date : 6/5/2013
//-----------------------------------------------------------------------------
void OutputMgr::outputHB(
  Parallel::Machine                     comm,
  const int                             stepNumber,
  const int                             maxStep,
  const Analysis::SweepVector &         step_sweep_parameters,
  const std::vector<double> &           timePoints,
  const std::vector<double> &           freqPoints,
  const Linear::BlockVector &           timeDomainSolutionVec,
  const Linear::BlockVector &           freqDomainSolutionVecReal,
  const Linear::BlockVector &           freqDomainSolutionVecImaginary,
  const Linear::BlockVector &           timeDomainLeadCurrentVec,
  const Linear::BlockVector &           freqDomainLeadCurrentVecReal,
  const Linear::BlockVector &           freqDomainLeadCurrentVecImaginary,
  const Linear::BlockVector &           timeDomainJunctionVoltageVec,
  const Linear::BlockVector &           freqDomainJunctionVoltageVecReal,
  const Linear::BlockVector &           freqDomainJunctionVoltageVecImaginary )
{
  // copy over the step sweep information:
  outputState_.stepLoopNumber_ = stepNumber;
  outputState_.stepMaxCount_ = maxStep;

  // copy the new values into the locally owned vector:
  if (!(step_sweep_parameters.empty()))
  {
    outputState_.stepSweepVector_ = step_sweep_parameters;
  }

  if (!activeOutputterStack_.empty())
  {
    std::vector<Outputter::Interface *>::const_iterator it = 
      activeOutputterStack_.back().begin();

    for (; it != activeOutputterStack_.back().end(); ++it)
    {
      (*it)->outputHB(comm, timePoints, freqPoints,
        timeDomainSolutionVec, 
        freqDomainSolutionVecReal, freqDomainSolutionVecImaginary,
        timeDomainLeadCurrentVec, 
        freqDomainLeadCurrentVecReal, freqDomainLeadCurrentVecImaginary,
        timeDomainJunctionVoltageVec, 
        freqDomainJunctionVoltageVecReal, freqDomainJunctionVoltageVecImaginary);
    }
  }
}

//-----------------------------------------------------------------------------
// Function      : OutputMgr::outputMPDE
// Purpose       : .PRINT output for mpde runs
// Special Notes :
// Scope         : public
// Creator       : Eric R. Keiter, SNL, Parallel Computational Sciences
// Creation Date : 9/22/03
//-----------------------------------------------------------------------------
void OutputMgr::outputMPDE(
  Parallel::Machine             comm,
  double                        time,
  const std::vector<double> &   fast_time_points,
  const Linear::BlockVector &   solution_vector)
{
  if (!activeOutputterStack_.empty())
  {
    std::vector<Outputter::Interface *>::const_iterator it = 
        activeOutputterStack_.back().begin(); 

    for ( ; it != activeOutputterStack_.back().end(); ++it)
    {
      (*it)->outputMPDE(comm, time, fast_time_points, solution_vector);
    }
  }
}

//-----------------------------------------------------------------------------
// Function      : OutputMgr::startStep
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Eric R. Keiter, SNL, Parallel Computational Sciences
// Creation Date : 08/30/04
//-----------------------------------------------------------------------------
void OutputMgr::startStep(int step, int max_step)
{
  if (!activeOutputterStack_.empty())
  {
    std::vector<Outputter::Interface *>::const_iterator it = 
      activeOutputterStack_.back().begin();

    for (;it != activeOutputterStack_.back().end(); ++it)
    {
      (*it)->startStep(step, max_step);
    }
  }
}

//-----------------------------------------------------------------------------
// Function      : OutputMgr::resetIndex
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Eric R. Keiter, SNL, Parallel Computational Sciences
// Creation Date : 08/30/04
//-----------------------------------------------------------------------------
void OutputMgr::resetIndex()
{
  if (!activeOutputterStack_.empty())
  {
    std::vector<Outputter::Interface *>::const_iterator it = 
      activeOutputterStack_.back().begin();

    for (; it != activeOutputterStack_.back().end(); ++it)
    {
      (*it)->resetIndex();
    }
  }
}

//-----------------------------------------------------------------------------
// Function      : OutputMgr::steppingComplete
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Eric R. Keiter, SNL, Parallel Computational Sciences
// Creation Date : 08/30/04
//-----------------------------------------------------------------------------
void OutputMgr::steppingComplete()
{
  if (!activeOutputterStack_.empty())
  {
    std::vector<Outputter::Interface *>::const_iterator it =
      activeOutputterStack_.back().begin();

    for (; it != activeOutputterStack_.back().end(); ++it)
    {
      (*it)->steppingComplete();
    }
  }
}

//-----------------------------------------------------------------------------
// Namespace     : Unnamed
// Purpose       : file-local scoped methods and data
// Special Notes :
// Creator       : dgbaur
// Creation Date : 06/28/2013
//-----------------------------------------------------------------------------
namespace {

//-----------------------------------------------------------------------------
// Function      : populateMetadata
// Purpose       : 
// Special Notes :
// Scope         : public
// Creator       : Dave Baur
// Creation Date : May 12, 2015
//-----------------------------------------------------------------------------
void populateMetadata(
  IO::PkgOptionsMgr &   options_manager)
{
  {
    Util::ParamMap &parameters = options_manager.addOptionsMetadataMap("OUTPUT");

    parameters.insert(Util::ParamMap::value_type("INITIAL_INTERVAL", Util::Param("INITIAL_INTERVAL", 0.0)));
    parameters.insert(Util::ParamMap::value_type("TIME", Util::Param("TIME", 0.0)));
    parameters.insert(Util::ParamMap::value_type("INTERVAL", Util::Param("INTERVAL", 0.0)));
    parameters.insert(Util::ParamMap::value_type("HDF5FILENAME", Util::Param("HDF5FILENAME", "")));
    parameters.insert(Util::ParamMap::value_type("PRINTHEADER", Util::Param("PRINTHEADER", true)));
    parameters.insert(Util::ParamMap::value_type("PRINTFOOTER", Util::Param("PRINTFOOTER", true)));
    parameters.insert(Util::ParamMap::value_type("OUTPUTVERSIONINRAWFILE", Util::Param("OUTPUTVERSIONINRAWFILE", false)));
  }

  {
    Util::ParamMap &parameters = options_manager.addOptionsMetadataMap("OUTPUT-LINE");

    parameters.insert(Util::ParamMap::value_type("TIME", Util::Param("TIME", 0.0)));
    parameters.insert(Util::ParamMap::value_type("INTERVAL", Util::Param("INTERVAL", 0.0)));
  }

  {
    Util::ParamMap &parameters = options_manager.addOptionsMetadataMap("PRINT");

    parameters.insert(Util::ParamMap::value_type("TYPE", Util::Param("TYPE", "TRAN")));
    parameters.insert(Util::ParamMap::value_type("FILE", Util::Param("FILE", "")));
    parameters.insert(Util::ParamMap::value_type("FORMAT", Util::Param("FORMAT", "STD")));
    parameters.insert(Util::ParamMap::value_type("DELIMITER", Util::Param("DELIMITER", "")));
    parameters.insert(Util::ParamMap::value_type("WIDTH", Util::Param("WIDTH", 17)));
    parameters.insert(Util::ParamMap::value_type("PRECISION", Util::Param("PRECISION", 8)));
    parameters.insert(Util::ParamMap::value_type("TIMESCALEFACTOR", Util::Param("TIMESCALEFACTOR", 1.0)));
    parameters.insert(Util::ParamMap::value_type("FILTER", Util::Param("FILTER", 0.0)));
  }
}

//-----------------------------------------------------------------------------
// Function      : extractPrintData
// Purpose       : Extract the parameters from a netlist .PRINT line held in
//                 parsed_line.
// Special Notes :
// Scope         : public
// Creator       : Lon Waters, SNL
// Creation Date : 05/02/2002
//-----------------------------------------------------------------------------
bool extractPrintData(
  IO::PkgOptionsMgr &           options_manager,
  IO::CircuitBlock &            circuit_block,
  const std::string &           netlist_filename,
  const IO::TokenVector &       parsed_line)
{
  Util::OptionBlock option_block("PRINT", 
      Util::OptionBlock::ALLOW_EXPRESSIONS, 
      netlist_filename, 
      parsed_line[0].lineNumber_);

  int numFields = parsed_line.size();

  if (DEBUG_IO) 
  {
    for (int ieric=0;ieric<parsed_line.size();++ieric)
    {
      Xyce::dout() << "parsed_line["<<ieric<<"] = " 
        << parsed_line[ieric].string_ << std::endl;
    }
  }

  // Set the TYPE and add it the parameters.
  std::string tmpString = parsed_line[1].string_;
  Util::toUpper(tmpString);
  if (tmpString == "TR")
  {
    tmpString = "TRAN"; // TR is a synonym for TRAN
  }
  Util::Param typeParameter("TYPE", tmpString);

  // Add the options parameter set to the model.
  addDefaultOptionsParameters(options_manager, option_block, "PRINT");

  // Reset the default TYPE with the value found.
  Util::Param *parameterPtr = Util::findParameter(option_block.begin(), option_block.end(), typeParameter.tag());
  if( parameterPtr == NULL )
  {
    Report::UserError0().at(netlist_filename, parsed_line[0].lineNumber_)
      << "Failed to find parameter\"" << typeParameter.usVal() << "\" in optionData for .PRINT statement";
  }

  parameterPtr->setVal( typeParameter.usVal() );

  // Set no default value for the output file.
  // if the user set FILE=vale it will be picked up here.  Otherwise
  // the output manager will use the top level simulation netlist name
  // as the default.
  Util::Param fileParameter("FILE", "");
  parameterPtr = Util::findParameter(option_block.begin(), option_block.end(), fileParameter.tag());

  if( parameterPtr == NULL )
  {
    Report::UserError0().at(netlist_filename, parsed_line[0].lineNumber_)
      << "Failed to find parameter\"" << typeParameter.usVal() << "\" in optionData for .PRINT statement";
  }

  parameterPtr->setVal( fileParameter.stringValue() );

  //
  // Note: The PRINT control parameters (FORMAT, WIDTH, FILE, ...) must be
  // tagged on the print line. Tagged parameters are "better" since they can
  // be identifed and parsed based soley on metadata with generic code.
  // The code that used to allow untagged parameters has been removed.

  // Check for tagged parameters.
  int parameterStartPos  = 2;
  int position = parameterStartPos;
  if ( numFields > parameterStartPos + 1 && parsed_line[parameterStartPos + 1].string_ == "=" )
  {
     // Tagged parameters found.
     Util::Param* parameterPtr;
     while ( position+1 < (int)parsed_line.size() &&
             parsed_line[position+1].string_ == "=" )
     {
       parameterPtr = Util::findParameter(option_block.begin(), option_block.end(), parsed_line[position].string_);
       if ( parameterPtr != NULL )
       {
         if (parameterPtr->tag() != "FILE")
           parameterPtr->setVal(std::string(ExtendedString(parsed_line[position+2].string_ ).toUpper()));
         else
           parameterPtr->setVal(std::string(ExtendedString(parsed_line[position+2].string_)));
       }
       else
       {
         Report::UserWarning0().at(netlist_filename, parsed_line[0].lineNumber_)
           << "No PRINT parameter " << parsed_line[position].string_ << " found, parameter will be ignored.";
         return false;
       }

       position += 3;
     }
  }
  else if (DEBUG_IO) // There is no point to this warning unless we're debugging
  {
    Report::UserWarning0().at(netlist_filename, parsed_line[0].lineNumber_)
      << "No tagged parameters found";
  }

  // Complete the PRINT line parsing.
  // Some of the remaining fields are of the
  // form I(Vname), V(node) or V(node1,node2), and these fields need
  // special treatment.
  Util::Param parameter;
  ExtendedString field("");
  std::ostringstream msg;
  int p_err=0;
  while ( position < numFields )
  {
    if (position+1 < numFields && parsed_line[position+1].string_ == "(")
    {
      if (parsed_line[position].string_ == "I" || parsed_line[position].string_ == "i"
        || ( (parsed_line[position].string_.size() == 2 || parsed_line[position].string_.size() == 3) &&
             (parsed_line[position].string_[0] == 'I' || parsed_line[position].string_[0] == 'i')))
      {
        if ((position+3 < numFields && parsed_line[position+3].string_ == ")") ||
            (position+4 < numFields && parsed_line[position+4].string_ == ")"))
        {
          field = parsed_line[position].string_;
          field.toUpper();
          parameter.setTag(field);
          parameter.setVal(1.0);
          option_block.addParam( parameter );

          if (parsed_line[position+3].string_ == ")")
          {
            field = parsed_line[position+2].string_;
            field.toUpper();
            parameter.setTag( field );
            parameter.setVal( 0.0 );
            option_block.addParam( parameter );

            position += 4;
          }
          else
          {
            field = parsed_line[position+2].string_ + " " + parsed_line[position+3].string_;
            field.toUpper();
            parameter.setTag( field );
            parameter.setVal( 0.0 );
            option_block.addParam( parameter );

            position += 5;
          }
        }
        else
        {
          msg << "Unrecognized current specification";
          p_err = position;
        }
      }
      else if( parsed_line[position].string_ == "V" || parsed_line[position].string_ == "v"
        || ( (parsed_line[position].string_.size() == 2 || parsed_line[position].string_.size() == 3) &&
             (parsed_line[position].string_[0] == 'V' || parsed_line[position].string_[0] == 'v')))
      {
        // position+3 < numFields test required to prevent a core dump on 
        // something invalid like .PRINT TRAN V(1
        if (position+3 < numFields && parsed_line[position+3].string_ == ")")
        {
          field = parsed_line[position].string_;
          field.toUpper();
          parameter.setTag(field);
          parameter.setVal( 1.0 );
          option_block.addParam( parameter );

          field = parsed_line[position+2].string_;
          field.toUpper();
          parameter.setTag( field );
          parameter.setVal( 0.0 );
          option_block.addParam( parameter );

          position += 4;
        }
        else if (position+5 < numFields && parsed_line[position+5].string_ == ")")
        {
          field = parsed_line[position].string_;
          field.toUpper();
          parameter.setTag(field);
          parameter.setVal( 2.0 );
          option_block.addParam( parameter );

          field = parsed_line[position+2].string_;
          field.toUpper();
          parameter.setTag( field );
          parameter.setVal( 0.0 );
          option_block.addParam( parameter );

          field = parsed_line[position+4].string_;
          field.toUpper();
          parameter.setTag( field );
          parameter.setVal( 0.0 );
          option_block.addParam( parameter );

          position += 6;
        }
        else
        {
          msg << "Unrecognized parenthetical specification";
          p_err = position;
        }
      }
      else if( parsed_line[position].string_ == "N" || parsed_line[position].string_ == "n")
      {
        if( position+3 < numFields && parsed_line[position+3].string_ == ")" )
        {
          parameter.setTag("N");
          parameter.setVal( 1.0 );
          option_block.addParam( parameter );

          field = parsed_line[position+2].string_;
          field.toUpper();
          parameter.setTag( field );
          parameter.setVal( 0.0 );
          option_block.addParam( parameter );

          position += 4;
        }
        else
        {
          msg << "Unrecognized parenthetical specification";
          p_err = position;
        }
      }
      else if( parsed_line[position].string_ == "W" || parsed_line[position].string_ == "w" ||
               parsed_line[position].string_ == "P" || parsed_line[position].string_ == "p")
      {
        if( position+3 < numFields && parsed_line[position+3].string_ == ")" )
        {
          field = parsed_line[position].string_;
          field.toUpper();
          parameter.setTag(field);
          //parameter.setTag("POWER_");
          parameter.setVal( 1.0 );
          option_block.addParam( parameter );

          field = parsed_line[position+2].string_;
          field.toUpper();
          parameter.setTag( field );
          parameter.setVal( 0.0 );
          option_block.addParam( parameter );

          position += 4;
        }
        else
        {
          msg << "Unrecognized parenthetical specification";
          p_err = position;
        }
      }
      else
      {
        msg << "Unrecognized parenthetical specification";
        p_err = position;
      }
    }
    else // This is either an expression, or a STEP parameter.
    {
      if (parsed_line[position].string_ == "(" || parsed_line[position].string_ == ")")
      {
        msg << "Unrecognized parenthesis";
        p_err = position;
      }
      field = parsed_line[position].string_;
      field.toUpper();
      parameter.setTag( field );
      parameter.setVal( 0.0 );
      option_block.addParam ( parameter );
      ++position;
    }

    if (!msg.str().empty())
    {
      msg << " in .print near ";
      position = p_err+4;
      p_err -= 2;
      if (p_err<0)
        p_err = 0;
      if (position >= numFields)
        position = numFields;
      while (p_err < position)
      {
        msg << parsed_line[p_err].string_ + " ";
        ++p_err;
      }
      Report::UserError0().at(netlist_filename, parsed_line[0].lineNumber_)
        << msg.str();
    }
  }

  circuit_block.addOptions(option_block);

  return true; // Only get here on success.
}

} // namespace <unnamed>

//-----------------------------------------------------------------------------
// Function      : registerPkgOptionsMgr
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Rich Schiek, 1437
// Creation Date : 10/21/08
//-----------------------------------------------------------------------------
bool registerPkgOptionsMgr(OutputMgr & output_manager, PkgOptionsMgr &options_manager)
{
  populateMetadata(options_manager);

  options_manager.addCommandParser(".PRINT", extractPrintData);

  //options_manager.addCommandProcessor("TRAN", 
  //IO::createRegistrationOptions(output_manager, &OutputMgr::registerTran));
  //options_manager.addCommandProcessor("TIMEINT", 
  //IO::createRegistrationOptions(output_manager, &OutputMgr::registerTranOptions));
//  options_manager.addCommandProcessor("MPDE", 
//  IO::createRegistrationOptions(output_manager, &OutputMgr::registerMPDETranOptions));
//  options_manager.addCommandProcessor("HB", 
//  IO::createRegistrationOptions(output_manager, &OutputMgr::registerHBOptions));
  options_manager.addCommandProcessor("PRINT", 
      IO::createRegistrationOptions(output_manager, &OutputMgr::parsePRINTBlock));
  options_manager.addCommandProcessor("SENS", 
      IO::createRegistrationOptions(output_manager, &OutputMgr::registerSens));
  options_manager.addCommandProcessor("NOISE", 
      IO::createRegistrationOptions(output_manager, &OutputMgr::registerNoise));

  options_manager.addOptionsProcessor("NONLIN", 
      IO::createRegistrationOptions(output_manager, &OutputMgr::registerNonlinearOptions));
  options_manager.addOptionsProcessor("OUTPUT", 
      IO::createRegistrationOptions(output_manager, &OutputMgr::registerOutputOptions));
//  options_manager.addOptionsProcessor("DEVICE", 
//  IO::createRegistrationOptions(output_manager, &OutputMgr::registerDeviceOptions));
  options_manager.addOptionsProcessor("SENSITIVITY", 
      IO::createRegistrationOptions(output_manager, &OutputMgr::registerSensOptions));

  return true;
}

} // namespace IO
} // namespace Xyce
