//-----------------------------------------------------------------------------
// Copyright Notice
//
//   Copyright 2002 Sandia Corporation. Under the terms
//   of Contract DE-AC04-94AL85000 with Sandia Corporation, the U.S.
//   Government retains certain rights in this software.
//
//    Xyce(TM) Parallel Electrical Simulator
//    Copyright (C) 2002-2016 Sandia Corporation
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
//
// Purpose        : Non-member functions that help the parser.
//
// Special Notes  :
//
// Creator        : Lon Waters, SNL
//
// Creation Date  : 09/06/2001
//
//
//
//
//-----------------------------------------------------------------------------

#include <Xyce_config.h>

#include <algorithm>
#include <iostream>
#include <fstream>
#include <functional>
#include <ctime>

#include <N_DEV_Configuration.h>
#include <N_ERH_ErrorMgr.h>
#include <N_IO_CircuitBlock.h>
#include <N_IO_CmdParse.h>
#include <N_IO_DeviceBlock.h>
#include <N_IO_ParameterBlock.h>
#include <N_IO_ParsingHelpers.h>
#include <N_UTL_ExtendedString.h>
#include <N_UTL_FeatureTest.h>
#include <N_PDS_MPI.h>
#include <N_PDS_Serial.h>

namespace Xyce {
namespace IO {


//--------------------------------------------------------------------------
// Function      : handleIncludeLine
// Purpose       : Handle a netlist .include or .lib line, add the include file
//                 to includeFiles_ for later processing.
// Special Notes :
// Creator       : Lon Waters
// Creation Date : 01/10/2001
//--------------------------------------------------------------------------
void handleIncludeLine(
    const bool firstLib,
    const std::string& netlistFileName,
    TokenVector const& parsedLine,
    const ExtendedString & ES1, std::string& includeFile, std::string& libSelect, std::string& libInside)
{
  bool lib = ES1.substr(0,4) != ".INC";

  std::string includeFileTmp;
  
  // Get the file name indicated by the .include line.
  if ( parsedLine.size() < 2 )
  {
    Report::UserWarning0().at(netlistFileName, parsedLine[0].lineNumber_)
      << ES1 << " is missing argument(s), ignoring";
    return;
  }
  else
  {
    includeFileTmp = parsedLine[1].string_;
  }

  if (!lib || (lib && (parsedLine.size() >= 3 || (parsedLine.size() >= 2 && firstLib)))) 
  {
    // Strip off the enclosing double quotes if they are present.
    if ( (includeFileTmp[0] == '"') &&
         (includeFileTmp[includeFileTmp.length()-1] == '"') )
    {
      includeFile = includeFileTmp.substr( 1, includeFileTmp.length()-2 );
    }
    else
    {
      includeFile = includeFileTmp;
    }
  }
  else
  {
    includeFile = "";
  }

  if ( lib )
  {
    if ( parsedLine.size() > 3)
    {
      Report::UserWarning0().at(netlistFileName, parsedLine[0].lineNumber_)
        << "Extraneous data on .LIB ignored";
    }
    if ( parsedLine.size() >= 3 )
    {
      libSelect = ExtendedString(parsedLine[2].string_).toUpper();
    }
    else
    {
      libInside = ExtendedString(parsedLine[1].string_).toUpper();
    }
  }
  else
  {
    if ( parsedLine.size() >= 3)
    {
      Report::UserWarning0().at(netlistFileName, parsedLine[0].lineNumber_)
        << "Extraneous data on .INCLUDE ignored";
    }
  }
}

//--------------------------------------------------------------------------
// Function      : handleEndlLine
// Purpose       : Handle a netlist .endl line
// Special Notes :
// Creator       : Dave Shirley, PSSI
// Creation Date : 10/09/2009
//--------------------------------------------------------------------------
void handleEndlLine(
    const std::string& netlistFileName,
    TokenVector const& parsedLine,
    const std::string& libSelect, std::string& libInside)
{
  if (libInside == "")
  {
    Report::UserError().at(netlistFileName, parsedLine[0].lineNumber_)
      << ".ENDL encountered without .LIB ";
    return;
  }

  if ( parsedLine.size() >= 2 )
  {
    ExtendedString libName ( parsedLine[1].string_ );
    libName.toUpper();
    if (libName != libInside)
    {
      Report::UserError().at(netlistFileName, parsedLine[0].lineNumber_)
        << ".ENDL encountered with name " << libName << ", which does not match .LIB name " << libInside;
    }
    if ( parsedLine.size() >2 )
    {
      Report::UserWarning().at(netlistFileName, parsedLine[0].lineNumber_)
        << "Extraneous field(s) following library name in .ENDL";
    }
  }
  else
  {
    Report::UserError().at(netlistFileName, parsedLine[0].lineNumber_)
      << ".ENDL encountered without library name, currently inside .LIB " << libInside;
  }

  libInside = "";
}

//--------------------------------------------------------------------------
// Function      : readLine
// Purpose       : Line-terminator-agnostic istream::getline() workalike for
//               : reading a single line from input stream
//
// Special Notes : file ptr is advanced
//               : line terminator is extracted
//--------------------------------------------------------------------------
void readLine( std::istream & in, std::string& line )
{
  line.clear();
  char c;

  // read all characters into line
  while( in.good() )
  {
    // read a char and append to line
    in.get( c );

    if( c == '\r' )
    {
      if( in.peek() == '\n' )
      {
        // extract the CR LF pair and discard
        in.get();
      }

      return;
    }
    else if ( in.eof() || c == '\n' )
    {
      return;
    }
    else
    {
      line.push_back( c );
    }
  }
}

//----------------------------------------------------------------------------
// Function       : splitLine
// Purpose        : Split an input character string into a tokenized vector.
// Special Notes  :
// Scope          : public
// Creator        : Heidi Thornquist
// Creation Date  : 10/15/2014
//----------------------------------------------------------------------------
void splitLine( const std::string& charLine, 
                TokenVector& line )
{
  int lineLength = charLine.length();
  int currPtr = 0;
  char c=charLine[ currPtr ];
  const std::string nonid(" \t\n\r\0");
  line.clear();

  while ( currPtr < lineLength )
  {
    StringToken field;
    field.string_.reserve(16);

    if (nonid.find(c) == nonid.npos) //not a whitespace character
    {
      field.string_ += c;
      c = charLine[ ++currPtr ];
      while ( currPtr < lineLength )
      {
        if (nonid.find(c) == nonid.npos)
        {
          field.string_ += c;
          c = charLine[ ++currPtr ];
        }
        else
        {
          break;
        }
      }
    }
    else
    {
      if (c == '\n' || c == '\r' || c == '\0')
      {
        field.string_ += c;
      }
      else
      {
        field.string_ += c;
        c = charLine[ ++currPtr ];
      }
    }

    if (field.string_.length() > 0)
    {
      line.push_back(field);
    }
  }
}

//--------------------------------------------------------------------------
// Function      : removeTwoTerminalDevice
// Purpose       : Given a two terminal device, this function checks to see
//                 if both nodes on the device are the same and, if so,
//                 decides whether or not the device should be removed
//                 from the circuit.  The decision to remove is based upon
//                 whether the specific device type was specified during the
//                 preprocessing phase as a device for which redundancies
//                 should be removed.  E.g., if the lines
//
//                 .PREPROCESS REMOVEUNUSED C
//                 C1 1 1 1
//                 R1 2 2 1
//
//                 appear in the netlist, the capacitor C1 will be removed
//                 from the netlist, whereas the resistor R1 will not.
//
// Special Notes : This function only determines *whether* a specific device
//                 should be removed; the process of removing the device from
//                 the netlist takes place in the handleLinePass1 and
//                 getLinePass2 functions.
//
// Creator       : Keith Santarelli
//
// Creation Date : 10/08/2007
//--------------------------------------------------------------------------
bool removeTwoTerminalDevice(const std::vector<bool>& pFilter,
                             const char& linetype,
                             const ExtendedString & node1, 
                             const ExtendedString & node2)
{
  bool result=false;

  if ( ( pFilter[PreprocessType::REDUNDANT_C] && linetype == 'C') 
     || ( pFilter[PreprocessType::REDUNDANT_D] && linetype == 'D') 
     || ( pFilter[PreprocessType::REDUNDANT_I] && linetype == 'I') 
     || ( pFilter[PreprocessType::REDUNDANT_L] && linetype == 'L') 
     || ( pFilter[PreprocessType::REDUNDANT_R] && linetype == 'R')
     || ( pFilter[PreprocessType::REDUNDANT_V] && linetype == 'V') )
  {
    if (node1 == node2)
    {
      result=true;
    }
  }

  return result;
}

//--------------------------------------------------------------------------
// Function      : removeThreeTerminalDevice
// Purpose       : Given a three terminal device, this function checks to see
//                 if all three nodes on the device are the same and, if so,
//                 decides whether or not the device should be removed
//                 from the circuit.  The decision to remove is based upon
//                 whether the specific device type was specified during the
//                 preprocessing phase as a device for which redundancies
//                 should be removed.  E.g., if the lines
//
//                 .PREPROCESS REMOVEUNUSED M
//                 M1 1 1 1 4 NMOS
//                 Q1 2 2 2 5 NPN
//
//                 appear in the netlist, the MOSFET M1 will be removed
//                 from the netlist, whereas the BJT Q1 will not.
//
// Special Notes : This function only determines *whether* a specific device
//                 should be removed; the process of removing the device from
//                 the netlist takes place in the handleLinePass1 and
//                 getLinePass2 functions.
//
// Creator       : Keith Santarelli
//
// Creation Date : 10/10/2007
//--------------------------------------------------------------------------
bool removeThreeTerminalDevice(const std::vector<bool>& pFilter,
                               const char& linetype,
                               const ExtendedString & node1,
                               const ExtendedString & node2,
                               const ExtendedString & node3)
{
  bool result=false;

  // Check three terminal devices.
  if ( (pFilter[PreprocessType::REDUNDANT_M] && linetype == 'M') 
        || (pFilter[PreprocessType::REDUNDANT_Q] && linetype == 'Q') )
  {
    if (node1 == node2 && node2 == node3)
    { 
      result=true;
    }
  }

  return result;
}


//--------------------------------------------------------------------------
// Function      : removeDevice
// Purpose       : Given a device block, this function checks to see
//                 if all the nodes on the device are the same and, if so,
//                 decides whether or not the device should be removed
//                 from the circuit.  The decision to remove is based upon
//                 whether the specific device type was specified during the
//                 preprocessing phase as a device for which redundancies
//                 should be removed.  E.g., if the lines
//
//                 .PREPROCESS REMOVEUNUSED M
//                 M1 1 1 1 4 NMOS
//                 Q1 2 2 2 5 NPN
//
//                 appear in the netlist, the MOSFET M1 will be removed
//                 from the netlist, whereas the BJT Q1 will not.
//
// Creator       : Heidi Thornquist
//
// Creation Date : 11/9/2015
//--------------------------------------------------------------------------
bool removeDevice(const std::vector<bool>& pFilter,
                  const DeviceBlock& device)
{ 
  bool result=false;
 
  // Get the vector of nodes from the device
  const std::vector<std::string> & nodes = device.getNodeValues();
  std::string linetype = device.getNetlistDeviceType();
  
  // Check two terminal devices.
  if ( ( pFilter[PreprocessType::REDUNDANT_C] && linetype == "C")
     || ( pFilter[PreprocessType::REDUNDANT_D] && linetype == "D")
     || ( pFilter[PreprocessType::REDUNDANT_I] && linetype == "I") 
     || ( pFilter[PreprocessType::REDUNDANT_L] && linetype == "L") 
     || ( pFilter[PreprocessType::REDUNDANT_R] && linetype == "R")
     || ( pFilter[PreprocessType::REDUNDANT_V] && linetype == "V") )
  { 
    if (nodes[0] == nodes[1])
    { 
      result=true;
    }
  }

  // Check three terminal devices.
  if ( (pFilter[PreprocessType::REDUNDANT_M] && linetype == "M") 
        || (pFilter[PreprocessType::REDUNDANT_Q] && linetype == "Q") )
  {
    if (nodes[0] == nodes[1] && nodes[1] == nodes[2])
    { 
      result=true;
    }
  }

  return result;
}



} // namespace IO
} // namespace Xyce
