//-----------------------------------------------------------------------------
// Copyright Notice
//
//   Copyright 2002 Sandia Corporation. Under the terms
//   of Contract DE-AC04-94AL85000 with Sandia Corporation, the U.S.
//   Government retains certain rights in this software.
//
//    Xyce(TM) Parallel Electrical Simulator
//    Copyright (C) 2002-2016 Sandia Corporation
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
//
// Purpose        : Specification file for the Abstract interface to sparse
//                  matrix type.
//
// Special Notes  :
//
// Creator        : Scott A. Hutchinson, SNL, Parallel Computational Sciences
//
// Creation Date  : 05/20/00
//
//
//
//
//-----------------------------------------------------------------------------

#ifndef Xyce_N_LAS_Matrix_h
#define Xyce_N_LAS_Matrix_h

// ---------- Standard Includes ----------
#include <string>
#include <vector>

// ----------   Xyce Includes   ----------
#include <N_LAS_fwd.h>
#include <N_NLS_fwd.h>


class N_PDS_ParMap;
class Epetra_CrsMatrix;
class Epetra_CrsGraph;

class Epetra_Export;
class Epetra_OffsetIndex;

namespace EpetraExt {

class CrsMatrix_View;

}

namespace Xyce {
namespace Linear {

//-----------------------------------------------------------------------------
// Class         : Matrix
// Purpose       : Abstract interface to sparse matrix type.
// Special Notes :
// Creator       : Scott A. Hutchinson, SNL, Parallel Compuational Sciences
// Creation Date : 05/20/00
//-----------------------------------------------------------------------------
class Matrix
{

public:

  //Constructors
    Matrix( N_PDS_ParMap & map, std::vector<int> & diagArray);

  Matrix( Epetra_CrsGraph * overlapGraph,
          Epetra_CrsGraph * baseGraph );

  //Constructor from an existing Epetra_CrsMatrix (makes copy of origMatrix)
  Matrix( Epetra_CrsMatrix * origMatrix, bool isOwned = true );

  //Destructor
  virtual ~Matrix();

  // This function needs to be invoked for a transpose solve.
  int setUseTranspose (bool useTranspose);
  bool useTranspose ();

  //Accumulate off processor fill contributions if necessary
  virtual void fillComplete();

  // Sparse-matrix vector multiply - multivector version.  If transA is true,
  // multiply by the transpose of matrix, otherwise just use matrix.
  void matvec(bool transA, const MultiVector & x, MultiVector & y);

  // Add in a matrix contribution
  void add( const Matrix & A );
  void addOverlap( const Matrix & A );

  void linearCombo ( const double a, const Matrix & A, 
                     const double b, const Matrix & B);

  //Filter to zero out values below threshold times inf norm
  void filterInfNorm(const double & threshold);
  //Filters to zero out values below threshold times the row sum
  void filterRowSum(const double & threshold);

  // Put function for the sparse-matrix.
  virtual void put(double s);

  // Scale the matrix
  void scale(double scaleFactor);

  // Diagonal Operations

  // Get the matrix diagonal (stored in an Vector)
  void getDiagonal(Vector & diagonal) const;
  // Sum a vector's values into the matrix diagonal
  bool sumIntoDiagonal(const Vector & vec);
  bool replaceDiagonal(const Vector & vec);

  // Row Operations

  // Get a row's maximum value
  double rowMax(int row) const;
  // Get a row's length (nonzero)
  int getRowLength(int row) const;
  // Get the non-zero values in a row
  void getRow(int row, int & length, double * coeffs, int * colIndices) const;

  // Get the non-zero values in a row
  void getRowCopy(int row, int length, int & numEntries, double * coeffs, int * colIndices) const;
  void getLocalRowCopy(int row, int length, int & numEntries, double * coeffs, int * colIndices) const;

  int getLocalNumRows() const;
  int getLocalRowLength(int row) const;
  //bool putLocalRow(int row, int length, double * coeffs, int * colIndices);
  bool shirleyPutRow(int row, int length, double * coeffs, int * colIndices);

  // Put a set of values into a row
  bool putRow(int row, int length, const double * coeffs, const int * colIndices);

  // Put a set of values into a row, using local indices
  bool putLocalRow(int row, int length, const double * coeffs, const int * colIndices);

  // Insert a set of values into a row
  void insertRow(int row, int length, double * coeffs, int * colIndices);

  // Replace a set of values into a row
  void replaceLocalRow(int row, int length, double * coeffs, int * colIndices);

  // Sum values into a row into the sparse matrix.
  bool sumIntoRow(int row, int length, const double * coeffs, const int * colIndices);

  // Sum values into a row into the sparse matrix, using local indices
  bool sumIntoLocalRow(int row, int length, const double * coeffs, const int * colIndices);

  // Return a pointer to a single row, col element.
  double * returnRawEntryPointer (int lidRow, int lidCol);

  // get a pointer to the compressed local row.
  int extractLocalRowView(int lidRow, int& numEntries, double*& values, int*& indices) const;

  // get a pointer to the compressed local row.
  int extractLocalRowView(int lidRow, int& numEntries, double*& values) const;

  // Tests for redundant indices first.
  bool sumIntoRowWithTest(int row, int length, const double * coeffs, 
      const int * colIndices);

#ifdef Xyce_LOCAL_INDEX_MATRIX
  // Sum values into a row into the sparse matrix.
  bool sumIntoLocalRow( int row,
                        int length,
                        const std::vector<double> & vals,
                        const std::vector<int> & colIndices);
#endif

  double * operator[]( int row );
  double * operator[]( int row ) const;

  Epetra_CrsMatrix & epetraObj() { return *aDCRSMatrix_; }
  const Epetra_CrsMatrix & epetraObj() const { return *aDCRSMatrix_; }

  // Output the matrix to a file
  void writeToFile(const char * filename, bool useLIDs = false, bool mmFormat=false );
  // Print the underlying Epetra objects
  virtual void printPetraObject(std::ostream &os) const;

  // Friend in the MultiVector and IterativeSolver classes so their
  // member functions can access our private members.
  friend class MultiVector;
  friend class FilteredMatrix;
  friend class Xyce::Nonlinear::DampedNewton;

protected:

  // Pointer the Petra multi-vector object.
  Epetra_CrsMatrix * aDCRSMatrix_;

  // Overlapped version of matrix
  Epetra_CrsMatrix * oDCRSMatrix_;

  // Subset View Transform
  EpetraExt::CrsMatrix_View * viewTransform_;

  // Importing Tools
  Epetra_Export * exporter_;
  Epetra_OffsetIndex * offsetIndex_;

private:

  // Default constructor (private)
  Matrix();

  // Copy constructor (private)
  Matrix(const Matrix & right);
  // Assignment operator (private)
  Matrix & operator = (const Matrix & right);

  bool operator == (const Matrix & right) const;
  bool operator != (const Matrix & right) const;

  // isOwned flag
  bool isOwned_;

  // Process library error codes.
  void processError(std::string methodMsg, int error) const;
};

} // namespace Linear
} // namespace Xyce

#endif
